@@ -400,7 +400,7 @@ impl Einsum {
         .unwrap()
     }
     #[staticmethod]
-    #[args(nodes = "*")]
+    #[args(nodes = "*", name = "None")]
     pub fn elementwise_broadcasted(nodes: Vec<CircuitRc>, name: Option<String>) -> Result<Einsum> {
         let out_shape = broadcast_shapes(
             &nodes
@@ -738,7 +738,7 @@ pub struct Rearrange {
     #[pyo3(get)]
     pub node: CircuitRc,
     #[pyo3(get)]
-    pub spec: RearrangeSpec,
+    pub spec: Box<RearrangeSpec>,
     info: CachedCircuitInfo,
     name: Option<String>,
 }
@@ -750,7 +750,7 @@ impl Rearrange {
         // TODO: reduce axes nums!
         let mut out = Self {
             node: node.clone(),
-            spec,
+            spec: Box::new(spec),
             name: Default::default(),
             info: Default::default(),
         };
@@ -796,7 +796,7 @@ impl Rearrange {
 
         if input_sizes.len() != node.info().rank() || any_sizes_dont_divide {
             return Err(ConstructError::RearrangeWrongInputShape {
-                spec: out.spec,
+                spec: (*out.spec).clone(),
                 shape: node.info().shape.clone(),
             }
             .into());
@@ -807,7 +807,7 @@ impl Rearrange {
     pub fn evolve(&self, node: Option<CircuitRc>, spec: Option<RearrangeSpec>) -> Rearrange {
         Rearrange::try_new(
             node.unwrap_or_else(|| self.node.clone()),
-            spec.unwrap_or_else(|| self.spec.clone()),
+            spec.unwrap_or_else(|| (*self.spec).clone()),
             self.name.clone(),
         )
         .unwrap()
@@ -874,7 +874,7 @@ impl CircuitNode for Rearrange {
     {
         Self::try_new(
             f(0, self.node.clone())?,
-            self.spec.clone(),
+            (*self.spec).clone(),
             self.name.clone(),
         )
     }
