@@ -1,7 +1,6 @@
 use anyhow::Result;
 use base16::encode_lower;
 use macro_rules_attribute::apply;
-use num_bigint::BigUint;
 use pyo3::{
     prelude::*,
     types::{IntoPyDict, PyTuple},
@@ -24,14 +23,6 @@ use crate::{
     CachedCircuitInfo, PyCircuitBase, Shape, TensorEvalError,
 };
 
-macro_rules! compute_info_auto_leaf_impl {
-    () => {
-        fn compute_max_non_leaf_size(&self) -> BigUint {
-            BigUint::from(0usize)
-        }
-    };
-}
-
 macro_rules! circuit_node_auto_leaf_impl {
     ($uuid:literal) => {
         circuit_node_auto_impl!($uuid);
@@ -73,7 +64,6 @@ impl CircuitNodeComputeInfoImpl for Array {
             TorchDeviceDtype::from_tensor(&self.value).into(),
         ))
     }
-    compute_info_auto_leaf_impl!();
 }
 
 impl CircuitNodeHashItems for Array {
@@ -173,7 +163,7 @@ impl Array {
 
             let mut kwargs = HashMap::default();
             if let Some(dtype) = device_dtype.dtype {
-                let dtype: &str = &dtype;
+                let dtype: &str = &String::from(&dtype);
                 kwargs.insert("dtype", PY_UTILS.torch.getattr(py, dtype).unwrap());
             }
             if let Some(device) = device_dtype.device {
@@ -240,7 +230,6 @@ impl CircuitNodeComputeInfoImpl for Symbol {
     fn compute_is_explicitly_computable(&self) -> bool {
         false
     }
-    compute_info_auto_leaf_impl!();
 }
 
 impl CircuitNodeHashItems for Symbol {
@@ -299,6 +288,12 @@ impl Symbol {
     pub fn new_with_none_uuid(shape: Shape, name: Option<String>) -> Self {
         Self::new(shape, Uuid::nil(), name)
     }
+
+    /// equality by ident is nice because we can guarantees about correct
+    /// behavior even when shapes change
+    pub fn ident(&self) -> (Uuid, Option<&str>) {
+        (self.uuid, self.name())
+    }
 }
 
 #[pyclass(extends=PyCircuitBase, unsendable)]
@@ -316,7 +311,6 @@ impl CircuitNodeComputeInfoImpl for Scalar {
     fn compute_shape(&self) -> Shape {
         self.info().shape.clone() // note: assumes shape has already been initted!
     }
-    compute_info_auto_leaf_impl!();
 }
 
 impl CircuitNodeHashItems for Scalar {
