@@ -1,26 +1,29 @@
 use std::{
     collections::BTreeMap,
     iter::{self, zip},
+    ops::Deref,
+    sync::Arc,
 };
 
 use anyhow::{bail, Context, Result};
 use macro_rules_attribute::apply;
 use pyo3::{exceptions::PyValueError, prelude::*};
 use rr_util::{
-    pycall, python_error_exception,
+    eq_by_big_hash::EqByBigHash,
+    impl_eq_by_big_hash, pycall, python_error_exception,
     rearrange_spec::{ExpandToSpecOrShape, OpShape, OpSize, RearrangeSpec},
     sv,
     symbolic_size::SymbolicSizeProduct,
-    tensor_util::{broadcast_shapes, Shape, TensorAxisIndex, TensorIndex},
-    util::EinsumAxes,
+    tensor_util::{right_align_shapes, Shape, TensorAxisIndex, TensorIndex},
+    util::{arc_unwrap_or_clone, EinsumAxes, HashBytes},
 };
 use rustc_hash::FxHashMap as HashMap;
 use smallvec::SmallVec;
 use thiserror::Error;
 
 use crate::{
-    deep_map_fallible_pre_new_children, prelude::*, Add, Concat, Einsum, GeneralFunction, Index,
-    Module, ModuleArgSpec, ModuleSpec, Rearrange, Scatter, SetSymbolicShape, Symbol, Tag,
+    deep_map_fallible_pre_new_children, prelude::*, Add, Concat, Cumulant, Einsum, GeneralFunction,
+    Index, Module, ModuleArgSpec, ModuleSpec, Rearrange, Scatter, SetSymbolicShape, Symbol, Tag,
 };
 
 #[pyfunction]
@@ -31,6 +34,51 @@ pub fn expand_node_py(circuit: CircuitRc, inputs: Vec<CircuitRc>) -> Result<Circ
     })
 }
 
+#[derive(Clone, Debug)]
+pub struct ReplaceMapRc {
+    replace_map: Arc<HashMap<CircuitRc, CircuitRc>>,
+    hash: HashBytes,
+}
+
+impl Default for ReplaceMapRc {
+    fn default() -> Self {
+        Self::new(Default::default())
+    }
+}
+
+impl Deref for ReplaceMapRc {
+    type Target = HashMap<CircuitRc, CircuitRc>;
+
+    fn deref(&self) -> &Self::Target {
+        &self.replace_map
+    }
+}
+
+impl ReplaceMapRc {
+    pub fn new(replace_map: HashMap<CircuitRc, CircuitRc>) -> Self {
+        let mut hasher = blake3::Hasher::new();
+        for (k, v) in replace_map.iter() {
+            hasher.update(&k.hash());
+            hasher.update(&v.hash());
+        }
+        Self {
+            replace_map: Arc::new(replace_map),
+            hash: hasher.finalize().into(),
+        }
+    }
+
+    pub fn into_map(self) -> <Self as Deref>::Target {
+        arc_unwrap_or_clone(self.replace_map)
+    }
+}
+
+impl EqByBigHash for ReplaceMapRc {
+    fn hash(&self) -> HashBytes {
+        self.hash
+    }
+}
+impl_eq_by_big_hash!(ReplaceMapRc);
+
 pub fn expand_node(
     circuit: CircuitRc,
     inputs: &Vec<CircuitRc>,
@@ -70,7 +118,7 @@ pub fn expand_node(
     // TODO: maybe we should allow for inconsistent symbolic batch shapes?
     // (probably not...)
     let broadcasted_batch_shape =
-        broadcast_shapes(&batch_shapes).context("batch shapes couldn't be broadcast together")?;
+        right_align_shapes(&batch_shapes).context("batch shapes didn't match (expand)")?;
     let batch_rank = broadcasted_batch_shape.len();
 
     let get_expanded_inputs = |batchable: &[bool]| -> Result<Vec<CircuitRc>, usize> {
@@ -112,7 +160,7 @@ pub fn expand_node(
 
     let out = match &**circuit {
         Circuit::SetSymbolicShape(_) => Ok(inputs[0].clone()), /* just return circuit, symbolic shapes will be added as needed! (TODO: maybe should keep?) */
-        Circuit::Symbol(_) | Circuit::Scalar(_) => Ok(circuit.clone()),
+        Circuit::Array(_) | Circuit::Symbol(_) | Circuit::Scalar(_) => Ok(circuit.clone()),
         Circuit::Rearrange(node) => {
             let input_shape_non_batch = inputs[0].info().shape[batch_ranks[0]..]
                 .iter()
@@ -261,6 +309,25 @@ pub fn expand_node(
             Concat::try_new(inputs, new_axis, node.name_cloned()).map(|x| x.rc())
         }
         Circuit::Add(node) => {
+            let inputs: Vec<_> = inputs
+                .iter()
+                .zip(&batch_ranks)
+                .map(|(inp, &b_rank)| {
+                    let unbatched_rank = inp.rank() - b_rank;
+                    let rank_diff = node.rank() - unbatched_rank;
+                    if rank_diff > 0 && b_rank > 0 {
+                        inp.unsqueeze(
+                            (b_rank..rank_diff + b_rank).collect(),
+                            inp.name().map(|x| format!("{} unsq_b_add", x)),
+                        )
+                        .unwrap()
+                        .rc()
+                    } else {
+                        inp.clone()
+                    }
+                })
+                .collect();
+
             let inputs = if let Some(max) = inputs.iter().map(|x| x.ndim()).max() {
                 let full_shape: OpShape = (0..max)
                     .map(|i| {
@@ -307,7 +374,7 @@ pub fn expand_node(
             // TODO: we could run into strange issues with general functions and expand.
             // Maybe resolve with an assert...
             // TODO: maybe alow for batching over individual inputs like:
-            // (This is a bit annoying in various ways and probably shouldn't be supported.
+            // (This is a bit annoying in various ways and probably shouldn't be supported).
             // - [b_0, b_1, b_2, (stuff)]
             // - [b_2, (other stuff)]
             GeneralFunction::try_new(
@@ -391,10 +458,21 @@ pub fn expand_node(
 
             Einsum::try_new(new_args, out_axes, node.name_cloned()).map(|x| x.rc())
         }
+        Circuit::Tag(node) => Ok(Tag::nrc(inputs[0].clone(), node.uuid, node.name_cloned())),
+        Circuit::Cumulant(node) => {
+            // For now, only supports expansion in easy cases
+            if zip(inputs, &node.nodes)
+                .skip(1)
+                .any(|(new_c, old_c)| new_c.rank() != old_c.rank())
+            {
+                bail!(ExpandError::CumulantRankChanged { circuit })
+            }
+            Ok(Cumulant::nrc(inputs.clone(), node.name_cloned()))
+        }
         Circuit::Module(node) => {
             assert_eq!(inputs.len(), node.children().count());
             let mut children = inputs.clone();
-            let rest = children.split_off(1);
+            let new_nodes = children.split_off(1);
             let spec_circuit = children.pop().unwrap();
 
             if node.nodes.is_empty() {
@@ -409,25 +487,6 @@ pub fn expand_node(
                 ));
             }
 
-            let new_nodes = rest
-                .chunks_exact(2)
-                .zip(&node.spec.arg_specs)
-                .map(|(sym_inp, arg_spec)| {
-                    if let [sym, inp] = sym_inp {
-                        if sym != &arg_spec.symbol.crc() {
-                            bail!(ExpandError::ModuleArgSpecSymbolChangedInExpand {
-                                old_symbol: arg_spec.symbol.clone(),
-                                new_symbol_circ: sym.clone(),
-                                module: node.clone(),
-                            });
-                        }
-                        Ok(inp.clone())
-                    } else {
-                        unreachable!()
-                    }
-                })
-                .collect::<Result<Vec<_>>>()?;
-
             // Approach:
             // suppose that orig node shapes are:
             // [     b_1, b_2, | x_0, x_1]
@@ -619,7 +678,6 @@ pub fn expand_node(
 
             Ok(final_out)
         }
-        Circuit::Tag(node) => Ok(Tag::new(inputs[0].clone(), node.uuid, node.name_cloned()).rc()),
         _ => {
             if inputs[..] == circuit.children().collect::<Vec<_>>()[..] {
                 Ok(circuit.clone())
@@ -678,6 +736,11 @@ pub enum ExpandError {
     #[error("trying to expand node, unknown variant {variant} ({e_name})")]
     NodeUnhandledVariant { variant: String },
 
+    #[error(
+        "Can only change the rank of the first circuit of the cumulant {circuit:?} ({e_name})"
+    )]
+    CumulantRankChanged { circuit: CircuitRc },
+
     #[error("Not currently supported! old_symbol={old_symbol:?} != new_symbol_circ={new_symbol_circ:?}\nmodule={module:?} ({e_name})")]
     ModuleArgSpecSymbolChangedInExpand {
         old_symbol: Symbol,
