@@ -391,7 +391,6 @@ impl SpecTrait for GeneralFunctionIndexSpec {
         Ok(Some(self.name()))
     }
 }
-
 #[derive(Clone)]
 pub struct PyWrap {
     ob: PyObject,
@@ -408,7 +407,7 @@ impl fmt::Debug for PyWrap {
     }
 }
 
-simple_from!(|x: PyWrap| -> GeneralFunctionSpec { GeneralFunctionSpec::Py(x) });
+simple_from!(|x: PyWrap| -> GeneralFunctionSpec { GeneralFunctionSpec::Py(Box::new(x)) });
 
 pub static PY_WRAP_BASE: GILLazyPy<PyObject> = GILLazyPy::new_py(|py| {
     let py_part = PyModule::from_code(
@@ -427,7 +426,7 @@ pub static PY_WRAP_BASE: GILLazyPy<PyObject> = GILLazyPy::new_py(|py| {
     get("GeneralFunctionSpecBase")
 });
 
-impl<'source> pyo3::FromPyObject<'source> for PyWrap {
+impl<'source> pyo3::FromPyObject<'source> for Box<PyWrap> {
     fn extract(ob: &'source PyAny) -> PyResult<Self> {
         let ob: PyObject = ob.into();
         let mut hasher = blake3::Hasher::new();
@@ -444,21 +443,21 @@ impl<'source> pyo3::FromPyObject<'source> for PyWrap {
         })?;
         hasher.update(&bytes);
 
-        Ok(Self {
+        Ok(Box::new(PyWrap {
             name: Python::with_gil(|py| -> Result<_> { Ok(ob.getattr(py, "name")?.extract(py)?) })?,
             ob,
             hash: *hasher.finalize().as_bytes(),
-        })
+        }))
     }
 }
 
-impl ToPyObject for PyWrap {
+impl ToPyObject for Box<PyWrap> {
     fn to_object(&self, _py: Python<'_>) -> PyObject {
         self.ob.clone()
     }
 }
 
-impl SpecTrait for PyWrap {
+impl SpecTrait for Box<PyWrap> {
     fn compute_hash(&self) -> HashBytes {
         self.hash
     }
@@ -500,7 +499,7 @@ impl SpecTrait for PyWrap {
 pub enum GeneralFunctionSpec {
     Simple(GeneralFunctionSimpleSpec),
     Index(GeneralFunctionIndexSpec),
-    Py(PyWrap),
+    Py(Box<PyWrap>), // boxed bc it was long pole for Circuit size
 }
 
 impl GeneralFunctionSpec {
