@@ -1,7 +1,7 @@
 #![feature(let_chains)]
 #![allow(clippy::borrow_deref_ref)]
 use std::{
-    collections::BTreeSet,
+    collections::{BTreeMap, BTreeSet},
     fmt::{self, Debug},
     hash::Hash,
     iter::zip,
@@ -50,6 +50,7 @@ use thiserror::Error;
 pub use variable_nodes::{DiscreteVar, StoredCumulantVar, Tag};
 
 use crate::{module::substitute_all_modules, print::print_circuit_stats};
+pub mod cached_circuit_properties;
 pub mod circuit_utils;
 pub mod computational_node;
 pub mod constant;
@@ -63,6 +64,7 @@ pub mod opaque_iterative_matcher;
 pub mod parsing;
 pub mod print;
 pub mod print_html;
+pub mod set_of_circuits;
 pub mod set_symbolic_size;
 pub mod variable_nodes;
 
@@ -356,7 +358,12 @@ pub trait CircuitNode: CircuitNodeInit + CircuitNodeSelfOnlyHash {
         ))
     }
 
-    fn mean(&self, axes: &[i64], name: Option<String>) -> Result<Einsum>
+    fn mean(
+        &self,
+        axes: &[i64],
+        name: Option<String>,
+        scalar_name: Option<String>,
+    ) -> Result<Einsum>
     where
         Self: Clone,
     {
@@ -364,8 +371,22 @@ pub trait CircuitNode: CircuitNodeInit + CircuitNodeSelfOnlyHash {
             .into_iter()
             .map(|i| self.info().shape[i as usize])
             .product();
-        self.sum(axes, name)?
-            .mul(Scalar::nrc(1. / (total_size as f64), sv![], None), None)
+        let sum = self.sum(axes, None)?;
+        Ok(Einsum::new(
+            sum.args
+                .into_iter()
+                .chain(std::iter::once((
+                    Scalar::nrc(
+                        1. / (total_size as f64),
+                        sv![],
+                        scalar_name.or_else(|| self.name().map(|s| format!("mean_div_{}", s))),
+                    ),
+                    sv![],
+                )))
+                .collect(),
+            sum.out_axes,
+            name,
+        ))
     }
 
     fn reduce(&self, op_name: String, axes: &[i64], name: Option<String>) -> Result<Circuit>
@@ -373,7 +394,7 @@ pub trait CircuitNode: CircuitNodeInit + CircuitNodeSelfOnlyHash {
         Self: Clone,
     {
         match op_name.as_str() {
-            "mean" => return self.mean(axes, name).map(CircuitNode::c),
+            "mean" => return self.mean(axes, name, None).map(CircuitNode::c),
             "sum" => return self.sum(axes, name).map(CircuitNode::c),
             _ => (),
         }
@@ -1188,7 +1209,6 @@ pub struct CachedCircuitInfo {
     pub is_explicitly_computable: bool,
     pub can_be_sampled: bool,
     pub hash: HashBytes,
-    pub max_non_leaf_size: BigUint,
     pub device_dtype: TorchDeviceDtypeOp,
     pub named_axes: NamedAxes,
     pub symbolic_size_constraints: SymbolicSizeConstraints,
@@ -1936,6 +1956,11 @@ impl PyCircuitBase {
         self.0.name().unwrap_or("")
     }
 
+    #[getter]
+    fn op_name(&self) -> Option<&str> {
+        self.0.name()
+    }
+
     #[getter]
     fn intermediate_cost_bound(&self) -> usize {
         self.0.intermediate_cost_bound()
@@ -1979,10 +2004,6 @@ impl PyCircuitBase {
         total_flops((*self.0).crc())
     }
 
-    pub fn max_non_input_size(&self) -> BigUint {
-        (*self.0).info().max_non_leaf_size.clone()
-    }
-
     pub fn print_stats(&self) {
         print_circuit_stats(&self.0)
     }
@@ -2042,6 +2063,10 @@ impl PyCircuitBase {
         total_arrayconstant_size(self.0.clone())
     }
 
+    fn get_device_dtype(&self) -> TorchDeviceDtypeOp {
+        self.0.info().device_dtype.clone()
+    }
+
     fn get_compatible_device_dtype(&self) -> TorchDeviceDtype {
         get_compatible_dtype(&***self)
     }
@@ -2077,10 +2102,17 @@ impl PyCircuitBase {
             .0
             .sum(&reduction_to_ints(axis, self.info().rank()), name)?)
     }
-    fn mean(&self, axis: Option<PyOpAtAxes>, name: Option<String>) -> Result<Einsum> {
-        Ok(self
-            .0
-            .mean(&reduction_to_ints(axis, self.info().rank()), name)?)
+    fn mean(
+        &self,
+        axis: Option<PyOpAtAxes>,
+        name: Option<String>,
+        scalar_name: Option<String>,
+    ) -> Result<Einsum> {
+        Ok(self.0.mean(
+            &reduction_to_ints(axis, self.info().rank()),
+            name,
+            scalar_name,
+        )?)
     }
     fn min(&self, axis: Option<PyOpAtAxes>, name: Option<String>) -> Result<CircuitRc> {
         Ok(self
@@ -2220,6 +2252,19 @@ impl PyCircuitBase {
         matcher.get_unique(self.0.clone(), fancy_validate)
     }
 
+    pub fn get_paths(
+        &self,
+        matcher: OpaqueIterativeMatcherVal,
+    ) -> PyResult<BTreeMap<CircuitRc, Vec<CircuitRc>>> {
+        matcher.get_paths(self.0.clone())
+    }
+    pub fn get_all_paths(
+        &self,
+        matcher: OpaqueIterativeMatcherVal,
+    ) -> PyResult<BTreeMap<CircuitRc, Vec<Vec<CircuitRc>>>> {
+        matcher.get_all_paths(self.0.clone())
+    }
+
     pub fn are_any_found(&self, matcher: OpaqueIterativeMatcherVal) -> PyResult<bool> {
         matcher.are_any_found(self.0.clone())
     }
