@@ -10,9 +10,11 @@ use rr_util::{
     lru_cache::TensorCacheRrfs,
     python_error_exception,
     symbolic_size::{SymbolicSizeProduct, SIZE_PROD_MATCH},
+    tensor_db::ensure_all_tensors_local,
     tensor_util::{
         parse_numeric, ParseError, PyParseError, Shape, TensorIndex, TorchDeviceDtypeOp,
     },
+    timed,
     util::{counts_g_1, flip_op_result, is_unique, NamedAxes},
 };
 use rustc_hash::FxHashMap as HashMap;
@@ -25,7 +27,6 @@ use crate::{
     ModuleArgSpec, ModuleSpec, Rearrange, Scalar, Scatter, SetSymbolicShape, StoredCumulantVar,
     Symbol, Tag,
 };
-
 const FANCY_PREFIX: &str = "fancy:";
 
 #[pyclass]
@@ -194,6 +195,7 @@ struct PartialCirc {
     pub name: Option<String>,
     pub children: Vec<CircuitIdent>,
     pub named_axes: Option<NamedAxes>,
+    pub module_arg_specs: Vec<ModuleArgSpec>,
 }
 
 impl Parser {
@@ -215,11 +217,25 @@ impl Parser {
         &self,
         string: &str,
         tensor_cache: &mut Option<TensorCacheRrfs>,
+    ) -> Result<Vec<CircuitRc>> {
+        self.parse_circuits_impl(
+            string,
+            &mut Default::default(),
+            &mut Default::default(),
+            tensor_cache,
+        )
+    }
+
+    fn parse_circuits_impl(
+        &self,
+        string: &str,
+        partial_circuits: &mut HashMap<CircuitIdent, PartialCirc>,
+        context: &mut HashMap<CircuitIdent, CircuitRc>,
+        tensor_cache: &mut Option<TensorCacheRrfs>,
     ) -> Result<Vec<CircuitRc>> {
         let lines: Vec<_> = string.lines().collect();
 
         let tab_width: usize = 2;
-        let mut partial_circuits: HashMap<CircuitIdent, PartialCirc> = HashMap::default();
         // ident and if it's a new node
         let mut stack: Vec<(CircuitIdent, bool)> = vec![];
         let mut was_previous_only_child_marker = false;
@@ -339,6 +355,14 @@ impl Parser {
 
             let is_new_node = !partial_circuits.contains_key(&circuit_ident);
 
+            let extra_and_parent_info =
+                re_captures.get(extra_cap).map(|z| z.as_str()).unwrap_or("");
+            let (extra, parent_info) = extra_and_parent_info
+                .split_once('!')
+                .map(|(x, y)| (x.trim(), Some(y.trim())))
+                .unwrap_or((extra_and_parent_info, None));
+            let extra = (!extra.is_empty()).then(|| extra.to_owned());
+
             let is_ref = circuit_ident
                 .as_str()
                 .map(|s| self.reference_circuits.contains_key(s))
@@ -347,7 +371,7 @@ impl Parser {
                 if re_captures.get(name_cap).is_some()
                     || re_captures.get(shape_cap).is_some()
                     || re_captures.get(variant_cap).is_some()
-                    || re_captures.get(extra_cap).is_some()
+                    || extra.is_some()
                 {
                     bail!(
                         ParseCircuitError::ReferenceCircuitNameFollowedByAdditionalInfo {
@@ -414,10 +438,6 @@ impl Parser {
                         .context("failed to parse variant in parse_circuit")
                 }))?;
 
-                let extra = re_captures
-                    .get(extra_cap)
-                    .and_then(|z| (!z.as_str().is_empty()).then(|| z.as_str().to_owned()));
-
                 if is_new_node {
                     partial_circuits.insert(
                         circuit_ident.clone(),
@@ -435,6 +455,7 @@ impl Parser {
                             extra: extra.unwrap_or_else(|| "".to_owned()),
                             children: vec![],
                             named_axes,
+                            module_arg_specs: vec![],
                         },
                     );
                 } else if self.on_repeat_check_info_same {
@@ -495,11 +516,21 @@ impl Parser {
                     });
                 }
 
-                partial_circuits
-                    .get_mut(l)
-                    .unwrap()
-                    .children
-                    .push(circuit_ident.clone());
+                self.add_child_info(
+                    l,
+                    &circuit_ident,
+                    parent_info,
+                    partial_circuits,
+                    context,
+                    tensor_cache,
+                )?;
+                let partial_parent = partial_circuits.get_mut(l).unwrap();
+                partial_parent.children.push(circuit_ident.clone());
+            } else if parent_info.is_some() {
+                bail!(ParseCircuitError::HasParentInfoButNoParent {
+                    ident: circuit_ident.clone(),
+                    line: get_unindented_line(),
+                });
             }
             if was_previous_only_child_marker {
                 stack.pop();
@@ -508,15 +539,142 @@ impl Parser {
             stack.push((circuit_ident, is_new_node));
         }
 
-        let mut context = HashMap::default();
+        let arrayconstant_tensor_hashes: Vec<String> = partial_circuits
+            .values()
+            .filter_map(|x| {
+                if !self.tensors_as_random && let CircuitType::Array = &x.variant && x.extra!="rand" {
+                    return Some(x.extra.clone());
+                }
+                None
+            })
+            .collect();
+        timed!(
+            ensure_all_tensors_local(arrayconstant_tensor_hashes),
+            100,
+            true
+        )?;
         top_level
             .iter()
             .map(|ident| {
-                self.deep_convert_partial_circ(ident, &partial_circuits, &mut context, tensor_cache)
+                self.deep_convert_partial_circ(ident, partial_circuits, context, tensor_cache)
             })
             .collect()
     }
 
+    fn add_module_child_info(
+        &self,
+        module_ident: &CircuitIdent,
+        child_ident: &CircuitIdent,
+        child_info: &str,
+        partial_circuits: &mut HashMap<CircuitIdent, PartialCirc>,
+        context: &mut HashMap<CircuitIdent, CircuitRc>,
+        tensor_cache: &mut Option<TensorCacheRrfs>,
+    ) -> Result<()> {
+        static RE_FLAGS: Lazy<Regex> = Lazy::new(|| Regex::new(&r"(.*) ([tf]{3})$").unwrap());
+
+        let (sym_str, flags_str) = RE_FLAGS
+            .captures(child_info)
+            .map(|x| (x.get(1).unwrap().as_str(), Some(x.get(2).unwrap().as_str())))
+            .unwrap_or((child_info, None));
+
+        let [batchable, expandable, ban_non_symbolic_size_expand] = flags_str
+            .map(|s| {
+                s.chars()
+                    .map(|x| TerseBool::try_from(x).unwrap().0)
+                    .collect::<Vec<_>>()
+                    .try_into()
+                    .unwrap()
+            })
+            .unwrap_or([true, true, true]);
+        // this is very obviously kind of a hack...
+        // not exactly sure what we should do instead.
+        let symbol_out_circs = self
+            .parse_circuits_impl(sym_str, partial_circuits, context, tensor_cache)
+            .with_context(|| {
+                format!(
+                    "{} child_ident={:?} {}",
+                    "failed to parse symbol circuit for module",
+                    child_ident,
+                    "extra info (perhaps flags are given incorrectly?)"
+                )
+            })?;
+        assert_eq!(symbol_out_circs.len(), 1);
+        let symbol = symbol_out_circs.into_iter().next().unwrap();
+        let symbol = symbol
+            .as_symbol()
+            .ok_or_else(|| ConstructError::ModuleExpectedSymbol {
+                actual_circuit: symbol.clone(),
+            })
+            .context("module expected symbol in parse (via child extra info)")?
+            .clone();
+
+        partial_circuits
+            .get_mut(module_ident)
+            .unwrap()
+            .module_arg_specs
+            .push(ModuleArgSpec {
+                symbol,
+                batchable,
+                expandable,
+                ban_non_symbolic_size_expand,
+            });
+
+        Ok(())
+    }
+
+    fn add_child_info(
+        &self,
+        parent_ident: &CircuitIdent,
+        child_ident: &CircuitIdent,
+        child_info: Option<&str>,
+        partial_circuits: &mut HashMap<CircuitIdent, PartialCirc>,
+        context: &mut HashMap<CircuitIdent, CircuitRc>,
+        tensor_cache: &mut Option<TensorCacheRrfs>,
+    ) -> Result<()> {
+        let parent = partial_circuits.get(parent_ident).unwrap();
+        let unexpected_child_info = || -> anyhow::Error {
+            ParseCircuitError::UnexpectedChildInfo {
+                child_info: child_info.unwrap().to_owned(),
+                parent_ident: parent_ident.clone(),
+                child_ident: child_ident.clone(),
+                parent_variant: parent.variant,
+            }
+            .into()
+        };
+
+        if parent.variant != CircuitType::Module {
+            if child_info.is_some() {
+                bail!(unexpected_child_info())
+            }
+            return Ok(());
+        }
+
+        if parent.children.is_empty() {
+            if child_info.is_some() {
+                bail!(unexpected_child_info().context("spec circuit had child info for module"))
+            }
+            return Ok(());
+        } else {
+            if child_info.is_none() {
+                bail!(
+                    ParseCircuitError::ModuleRequiresInputChildToHaveParentInfo {
+                        module_ident: parent_ident.clone(),
+                        child_ident: child_ident.clone(),
+                    }
+                )
+            }
+        }
+
+        self.add_module_child_info(
+            parent_ident,
+            child_ident,
+            child_info.unwrap(),
+            partial_circuits,
+            context,
+            tensor_cache,
+        )
+    }
+
     fn deep_convert_partial_circ(
         &self,
         ident: &CircuitIdent,
@@ -545,10 +703,14 @@ impl Parser {
             .deep_convert_partial_circ_children(ident, ps, children, tensor_cache)
             .with_context(|| {
                 format!(
-                    "in parse, failed to convert ident={:?}, variant={}\nfull partial circuit: {:?}",
+                    concat!(
+                        "in parse, failed to convert ident={:?}, ",
+                        "variant={}\nfull partial circuit: {:?}"
+                    ),
                     ident, ps.variant, ps
                 )
-            })?.rename(partial_circuits[ident].name.clone());
+            })?
+            .rename(partial_circuits[ident].name.clone());
 
         context.insert(ident.clone(), result.clone());
         Ok(result)
@@ -574,14 +736,20 @@ impl Parser {
 
         let extra_should_be_empty = || {
             if !ps.extra.is_empty() {
-                bail!(ParseError::ExpectedNoExtraInfo {
-                    extra: ps.extra.clone()
+                bail!(ParseCircuitError::ExpectedNoExtraInfo {
+                    extra: ps.extra.clone(),
+                    variant: ps.variant
                 })
             }
             Ok(())
         };
 
         type T = CircuitType;
+
+        if ps.variant != T::Module {
+            assert!(ps.module_arg_specs.is_empty());
+        }
+
         let result = match ps.variant {
             T::Array => {
                 expected_k_children(0)?;
@@ -644,12 +812,17 @@ impl Parser {
                     &ps.extra[FANCY_PREFIX.len()..],
                     children,
                     ps.name.clone(),
-                ).context(
-                    "Couldn't parse fancy einsum string. If you wanted to parse a regular einsum, remove \"fancy:\".",
                 )
+                .context(concat!(
+                    "Couldn't parse fancy einsum string. If you wanted ",
+                    "to parse a regular einsum, remove \"fancy:\"."
+                ))
             } else {
                 Einsum::from_einsum_string(&ps.extra, children.clone(), ps.name.clone()).context(
-                    "Couldn't parse einsum string. If you wanted to parse a fancy einsum string, prefix it by \"fancy:\".",
+                    concat!(
+                        "Couldn't parse einsum string. If you wanted ",
+                        "to parse a fancy einsum string, prefix it by \"fancy:\"."
+                    ),
                 )
             }?
             .rc(),
@@ -709,85 +882,17 @@ impl Parser {
                 .rc()
             }
             T::Module => {
-                let arg_info = if ps.extra == "all_t" {
-                    if children.len() % 2 != 1 {
-                        bail!(ParseCircuitError::ModuleInconsistentNumChildren {
-                            name: ps.name.clone(),
-                            extra: ps.extra.clone(),
-                            num_children: children.len(),
-                            num_args: None
-                        })
-                    }
-                    (0..((children.len() - 1) / 2))
-                        .map(|_| [true, true, true])
-                        .collect()
-                } else {
-                    ps.extra
-                        .split(",")
-                        .map(|x| {
-                            let num_chars = x.chars().count();
-                            if num_chars != 3 {
-                                bail!(ParseCircuitError::Expected3CharsPerForModuleExtra {
-                                    name: ps.name.clone(),
-                                    num_chars,
-                                    extra: ps.extra.clone(),
-                                });
-                            }
-                            let out = x
-                                .chars()
-                                .map(|x| TerseBool::try_from(x).map(|b| b.0))
-                                .collect::<Result<Vec<_>>>()?
-                                .try_into()
-                                .unwrap();
-                            Ok(out)
-                        })
-                        .collect::<Result<Vec<[bool; 3]>>>()?
-                };
-
-                if children.len() != 2 * arg_info.len() + 1 {
-                    bail!(ParseCircuitError::ModuleInconsistentNumChildren {
+                extra_should_be_empty()?;
+                if children.len() == 0 {
+                    bail!(ParseCircuitError::ModuleNoSpecCircuit {
                         name: ps.name.clone(),
-                        extra: ps.extra.clone(),
-                        num_children: children.len(),
-                        num_args: Some(arg_info.len())
                     })
                 }
 
-                let rest = children.split_off(1);
+                let nodes = children.split_off(1);
                 let spec_circuit = children.pop().unwrap();
-
-                let (arg_specs, nodes) = rest
-                    .chunks_exact(2)
-                    .zip(arg_info)
-                    .map(
-                        |(sym_inp, [batchable, expandable, ban_non_symbolic_size_expand])| {
-                            let (sym, inp) = if let [sym, inp] = sym_inp {
-                                (sym, inp)
-                            } else {
-                                unreachable!()
-                            };
-                            let symbol = sym
-                                .as_symbol()
-                                .ok_or_else(|| ConstructError::ModuleExpectedSymbol {
-                                    actual_circuit: sym.clone(),
-                                })
-                                .context("module expected symbol in parse")?
-                                .clone();
-
-                            Ok((
-                                ModuleArgSpec {
-                                    symbol,
-                                    batchable,
-                                    expandable,
-                                    ban_non_symbolic_size_expand,
-                                },
-                                inp.clone(),
-                            ))
-                        },
-                    )
-                    .collect::<Result<Vec<_>>>()?
-                    .into_iter()
-                    .unzip();
+                let arg_specs = ps.module_arg_specs.clone();
+                assert_eq!(arg_specs.len(), nodes.len());
 
                 Module::try_new(
                     nodes,
@@ -975,6 +1080,14 @@ pub enum ParseCircuitError {
         ident: CircuitIdent,
     },
 
+    #[error("child_info={child_info:?} for child_ident={child_ident:?}, parent_ident={parent_ident:?}, and parent_variant={parent_variant}, but expected None ({e_name})")]
+    UnexpectedChildInfo {
+        child_info: String,
+        child_ident: CircuitIdent,
+        parent_ident: CircuitIdent,
+        parent_variant: CircuitType,
+    },
+
     #[error("Parsing invalid circuit variant {v} ({e_name})")]
     InvalidVariant { v: String },
 
@@ -999,6 +1112,9 @@ pub enum ParseCircuitError {
         child_line: String,
     },
 
+    #[error("ident={ident:?} line='{line}' ({e_name})")]
+    HasParentInfoButNoParent { ident: CircuitIdent, line: String },
+
     #[error("reference_name={reference_name} ({e_name})")]
     ReferenceCircuitNameFollowedByAdditionalInfo {
         reference_name: String,
@@ -1028,22 +1144,20 @@ pub enum ParseCircuitError {
         err_strs: String,
     },
 
-    #[error("num_chars={num_chars}!=3 extra='{extra}' name={name:?}\nmodule extra should look like 'ttt,tft,ftf' and we check for 3 chars per comma delimited item.\n({e_name})")]
-    Expected3CharsPerForModuleExtra {
-        name: Option<String>,
-        num_chars: usize,
-        extra: String,
-    },
+    #[error(
+        "Module has no children, but should have at least a spec circuit name={name:?} ({e_name})"
+    )]
+    ModuleNoSpecCircuit { name: Option<String> },
 
-    #[error("num_children={num_children} != 2 * num_args={num_args:?} + 1\nname={name:?} extra={extra}\n{}\n{}\n{}\n({e_name})",
-    "modules have 1 child for the spec and then a symbol and a input circuit for each arg (2 * num_args + 1)",
-    "this implies the number of children is always odd",
-    "num_args is based on parsing the batchable and expandable info from 'extra' (see above) (in a form like 'ttt,tft,ftf' (Some(3)) or 'all_t' (None))")]
-    ModuleInconsistentNumChildren {
-        name: Option<String>,
-        num_children: usize,
-        extra: String,
-        num_args: Option<usize>,
+    #[error(
+        "for module_ident={module_ident:?} the input child_ident={child_ident:?} {}\n{}\n{}\n({e_name})",
+        "doesn't have any extra parent info.",
+        "modules must indicate which symbol is replaced for each input.",
+        "this is done with '!', as in: \"'my_input' Einsum a->a ! 'my_sym'\""
+    )]
+    ModuleRequiresInputChildToHaveParentInfo {
+        module_ident: CircuitIdent,
+        child_ident: CircuitIdent,
     },
 
     #[error("extra='{extra}' invalid. Should contain | (TODO: example). ({e_name})")]
