@@ -2,6 +2,7 @@ use std::{
     fmt,
     fmt::{Debug, Write},
     hash::{Hash, Hasher},
+    iter,
     ops::Deref,
     sync::Mutex,
 };
@@ -18,14 +19,16 @@ use rr_util::{
 use rustc_hash::{FxHashMap as HashMap, FxHashSet as HashSet};
 
 use crate::{
+    cached_circuit_properties::max_non_leaf_size,
     circuit_utils::{count_nodes, total_flops},
     generalfunction::SpecTrait,
-    opaque_iterative_matcher::{OpaqueIterativeMatcher, OpaqueIterativeMatcherVal},
+    opaque_iterative_matcher::{
+        get_opaque_type_matcher, OpaqueIterativeMatcher, OpaqueIterativeMatcherVal,
+    },
     parsing::{get_reference_circuits, ParseCircuitError},
     prelude::*,
-    HashBytes,
+    CircuitType, HashBytes,
 };
-
 // TODO: improve color scheme
 const COLOR_CODES: [usize; 14] = [31, 32, 33, 34, 35, 36, 90, 91, 92, 93, 94, 95, 96, 97];
 const NAME_TO_COLOR: Lazy<HashMap<String, usize>> = Lazy::new(|| {
@@ -136,6 +139,8 @@ pub struct PrintOptions {
     #[pyo3(get, set)]
     pub force_use_serial_numbers: bool,
     #[pyo3(get, set)]
+    pub number_leaves: bool,
+    #[pyo3(get, set)]
     pub colorer: Option<CircuitToColorCode>,
     #[pyo3(get, set)]
     pub comment_arg_names: bool,
@@ -149,6 +154,8 @@ pub struct PrintOptions {
     pub only_child_below: bool,
     #[pyo3(get, set)]
     pub tensor_index_literal: bool,
+    #[pyo3(get, set)]
+    pub show_all_named_axes: bool,
 }
 
 impl Default for PrintOptions {
@@ -162,6 +169,7 @@ impl Default for PrintOptions {
             arrows: false,
             print_size_threshold: None,
             force_use_serial_numbers: false,
+            number_leaves: false,
             colorer: None,
             comment_arg_names: false,
             commenters: vec![],
@@ -169,6 +177,7 @@ impl Default for PrintOptions {
             seen_children_same_line: false,
             only_child_below: false,
             tensor_index_literal: false,
+            show_all_named_axes: false,
         }
         .validate_ret()
         .unwrap()
@@ -192,6 +201,118 @@ impl PrintOptions {
         self.validate()?;
         Ok(self)
     }
+
+    /// returns whether or not we finished here
+    pub fn repr_full_line(
+        &self,
+        circ: &Circuit,
+        result: &mut String,
+        seen_hashes: &mut HashMap<HashBytes, (String, String)>,
+        runnning_serial_number: &mut usize,
+        disallow_name_ident: Option<&HashMap<String, bool>>,
+    ) -> Result<bool> {
+        if let Some(ident) = self.reference_circuits.get(&circ.clone().rc()) {
+            result.push_str(&self.string_escape(ident));
+            return Ok(true);
+        }
+
+        let mut repeat_info = None; // only matters for 'print_as_tree';
+        if let Some((rep_str, line_prefix)) = seen_hashes.get(&circ.info().hash) {
+            if !self.is_print_as_tree() {
+                result.push_str(rep_str);
+                return Ok(true);
+            } else {
+                repeat_info = Some((rep_str.clone(), line_prefix.clone()));
+            }
+        }
+
+        // allow_name_ident is_some iff !self.force_use_serial_numbers
+        let disallow_name = disallow_name_ident
+            .zip(circ.name())
+            .map(|(map, name)| map[name])
+            .unwrap_or(true);
+
+        let (repeat_string, line_prefix) = match (repeat_info, disallow_name, circ.name()) {
+            (Some(x), _, _) => x,
+            (None, true, Some(name)) => {
+                let escaped = self.string_escape(name);
+                let this_serial_number = runnning_serial_number.to_string();
+                *runnning_serial_number += 1;
+                let out = this_serial_number + " " + &escaped;
+                (out.clone(), out)
+            }
+            (None, false, Some(name)) => {
+                let escaped = self.string_escape(name);
+                (escaped.clone(), escaped)
+            }
+            (None, true, None) => {
+                let this_serial_number = runnning_serial_number.to_string();
+                *runnning_serial_number += 1;
+                (
+                    this_serial_number.clone() + " " + &circ.variant_string(),
+                    this_serial_number,
+                )
+            }
+
+            (None, false, None) => unreachable!(),
+        };
+
+        result.push_str(&line_prefix);
+
+        seen_hashes.insert(circ.info().hash, (repeat_string, line_prefix));
+
+        result.push_str(
+            &self
+                .repr_line_info(circ.clone().rc())
+                .with_context(|| format!("repr line info failed for circuit={:?}", circ))?,
+        );
+
+        Ok(false)
+    }
+
+    pub fn has_child_info(&self, circ: &Circuit) -> bool {
+        circ.is_module() && circ.num_children() > 0
+    }
+
+    pub fn get_child_info(
+        &self,
+        circ: &Circuit,
+        seen_hashes: &mut HashMap<HashBytes, (String, String)>,
+        runnning_serial_number: &mut usize,
+        disallow_name_ident: Option<&HashMap<String, bool>>,
+        child_idx: usize,
+    ) -> Option<String> {
+        if let Some(m) = circ.as_module() {
+            if child_idx == 0 {
+                return None;
+            }
+
+            let arg_spec = &m.spec.arg_specs[child_idx - 1];
+            let mut result = String::new();
+            self.repr_full_line(
+                &arg_spec.symbol.clone().c(),
+                &mut result,
+                seen_hashes,
+                runnning_serial_number,
+                disallow_name_ident,
+            )
+            .unwrap();
+
+            if !(arg_spec.batchable && arg_spec.expandable && arg_spec.ban_non_symbolic_size_expand)
+            {
+                result.push_str(&format!(
+                    " {}{}{}",
+                    TerseBool(arg_spec.batchable),
+                    TerseBool(arg_spec.expandable),
+                    TerseBool(arg_spec.ban_non_symbolic_size_expand)
+                ));
+            }
+
+            Some(result)
+        } else {
+            None
+        }
+    }
 }
 
 const DEFAULT_END_DEPTH: usize = 2;
@@ -268,13 +389,15 @@ impl PrintOptions {
         arrows = "PrintOptions::default().arrows",
         print_size_threshold = "None",
         force_use_serial_numbers = "PrintOptions::default().force_use_serial_numbers",
+        number_leaves = "PrintOptions::default().number_leaves",
         colorer = "None",
         comment_arg_names = "PrintOptions::default().comment_arg_names",
         commenters = "PrintOptions::default().commenters",
         sync_tensors = "PrintOptions::default().sync_tensors",
         seen_children_same_line = "PrintOptions::default().seen_children_same_line",
         only_child_below = "PrintOptions::default().only_child_below",
-        tensor_index_literal = "PrintOptions::default().tensor_index_literal"
+        tensor_index_literal = "PrintOptions::default().tensor_index_literal",
+        show_all_named_axes = "PrintOptions::default().show_all_named_axes"
     )]
     pub fn new(
         bijection: bool,
@@ -286,6 +409,7 @@ impl PrintOptions {
         arrows: bool,
         print_size_threshold: Option<usize>,
         force_use_serial_numbers: bool,
+        number_leaves: bool,
         colorer: Option<CircuitToColorCode>,
         comment_arg_names: bool,
         commenters: Vec<CircuitCommenter>,
@@ -293,6 +417,7 @@ impl PrintOptions {
         seen_children_same_line: bool,
         only_child_below: bool,
         tensor_index_literal: bool,
+        show_all_named_axes: bool,
     ) -> Result<Self> {
         let result = Self {
             bijection,
@@ -312,6 +437,7 @@ impl PrintOptions {
             arrows,
             print_size_threshold,
             force_use_serial_numbers,
+            number_leaves,
             colorer,
             comment_arg_names,
             commenters,
@@ -319,6 +445,7 @@ impl PrintOptions {
             seen_children_same_line,
             only_child_below,
             tensor_index_literal,
+            show_all_named_axes,
         };
         result.validate()?;
         Ok(result)
@@ -333,13 +460,15 @@ impl PrintOptions {
         arrows = "Default::default()",
         print_size_threshold = "Default::default()",
         force_use_serial_numbers = "Default::default()",
+        number_leaves = "Default::default()",
         colorer = "Default::default()",
         comment_arg_names = "Default::default()",
         commenters = "Default::default()",
         sync_tensors = "Default::default()",
         seen_children_same_line = "Default::default()",
         only_child_below = "Default::default()",
-        tensor_index_literal = "Default::default()"
+        tensor_index_literal = "Default::default()",
+        show_all_named_axes = "Default::default()"
     )]
     pub fn evolve(
         &self,
@@ -351,6 +480,7 @@ impl PrintOptions {
         arrows: MaybeNotSet<bool>,
         print_size_threshold: MaybeNotSet<Option<usize>>,
         force_use_serial_numbers: MaybeNotSet<bool>,
+        number_leaves: MaybeNotSet<bool>,
         colorer: MaybeNotSet<Option<CircuitToColorCode>>,
         comment_arg_names: MaybeNotSet<bool>,
         commenters: MaybeNotSet<Vec<CircuitCommenter>>,
@@ -358,6 +488,7 @@ impl PrintOptions {
         seen_children_same_line: MaybeNotSet<bool>,
         only_child_below: MaybeNotSet<bool>,
         tensor_index_literal: MaybeNotSet<bool>,
+        show_all_named_axes: MaybeNotSet<bool>,
     ) -> Result<Self> {
         let cloned = self.clone();
         Self::new(
@@ -376,6 +507,7 @@ impl PrintOptions {
             force_use_serial_numbers
                 .0
                 .unwrap_or(cloned.force_use_serial_numbers),
+            number_leaves.0.unwrap_or(cloned.number_leaves),
             colorer.0.unwrap_or(cloned.colorer),
             comment_arg_names.0.unwrap_or(cloned.comment_arg_names),
             commenters.0.unwrap_or(cloned.commenters),
@@ -387,6 +519,7 @@ impl PrintOptions {
             tensor_index_literal
                 .0
                 .unwrap_or(cloned.tensor_index_literal),
+            show_all_named_axes.0.unwrap_or(cloned.show_all_named_axes),
         )
     }
 
@@ -424,27 +557,59 @@ impl PrintOptions {
         .unwrap()
     }
 
+    #[staticmethod]
+    pub fn type_nest_default(
+        circuit_type: CircuitType,
+        traversal: Option<OpaqueIterativeMatcherVal>,
+    ) -> PrintOptions {
+        Self {
+            bijection: false,
+            number_leaves: true,
+            traversal: Some(get_opaque_type_matcher(circuit_type, traversal)),
+            ..Default::default()
+        }
+        .validate_ret()
+        .unwrap()
+    }
+
+    #[staticmethod]
+    pub fn einsum_nest_default(traversal: Option<OpaqueIterativeMatcherVal>) -> PrintOptions {
+        Self::type_nest_default(CircuitType::Einsum, traversal)
+    }
+
+    #[staticmethod]
+    pub fn add_nest_default(traversal: Option<OpaqueIterativeMatcherVal>) -> PrintOptions {
+        Self::type_nest_default(CircuitType::Add, traversal)
+    }
+
+    #[getter]
+    pub fn is_print_as_tree(&self) -> bool {
+        // other configuration might require printing as tree
+        self.number_leaves
+    }
+
     pub fn validate(&self) -> Result<()> {
         if self.bijection && self.leaves_on_top {
-            bail!(anyhow!("bijection print cant have leaves on top"))
+            bail!("bijection print cant have leaves on top")
         }
         if self.bijection && self.traversal.is_some() {
-            bail!(anyhow!(
-                "bijection print cant terminate early, so traversal must be None"
-            ))
+            bail!("bijection print cant terminate early, so traversal must be None")
         }
         if self.bijection && self.seen_children_same_line {
-            bail!(anyhow!(
-                "bijection doesn't support seen children same line, but technically could"
-            ))
+            bail!("bijection doesn't support seen children same line, but technically could")
         }
         if self.bijection && self.tensor_index_literal {
-            bail!(anyhow!("bijection doesn't support index tensor literal"))
+            bail!("bijection doesn't support index tensor literal")
         }
-        if self.seen_children_same_line && self.comment_arg_names {
-            bail!(anyhow!(
-                "children same line + comment arg names not supported"
-            ))
+        if self.seen_children_same_line
+            && (self.comment_arg_names
+                || !self.commenters.is_empty()
+                || self.print_size_threshold.is_some())
+        {
+            bail!("children same line + commenting (arg names, commenter, print_size_threshold) not supported")
+        }
+        if self.bijection && self.is_print_as_tree() {
+            bail!("bijection print can't print as tree (caused by number_leaves)")
         }
         Ok(())
     }
@@ -476,14 +641,18 @@ impl PrintOptions {
                     | Circuit::Array(_)
                     | Circuit::SetSymbolicShape(_)
             )
-            || !circ.info().named_axes.is_empty()
+            || self.show_all_named_axes && !circ.info().named_axes.is_empty()
         {
             result.push_str(&format!(
                 " [{}]",
                 (0..circ.info().rank())
                     .map(|x| match circ.info().named_axes.get(&(x as u8)) {
                         None => SymbolicSizeProduct::from(circ.shape()[x]).to_string(),
-                        Some(s) => format!("{}:{}", s, circ.shape()[x]),
+                        Some(s) => format!(
+                            "{}:{}",
+                            s,
+                            SymbolicSizeProduct::from(circ.shape()[x]).to_string()
+                        ),
                     })
                     .collect::<Vec<_>>()
                     .join(",")
@@ -544,33 +713,6 @@ impl PrintOptions {
                         format!("{}", &sy.uuid)
                     }
                 }
-                Circuit::Module(mn) => {
-                    if mn.spec.arg_specs.len() >= 2
-                        && mn
-                            .spec
-                            .arg_specs
-                            .iter()
-                            .all(|x| x.batchable && x.expandable && x.ban_non_symbolic_size_expand)
-                    {
-                        "all_t".to_owned()
-                    } else {
-                        // print out out arg info, everything else in children
-                        // TODO: less terse?
-                        mn.spec
-                            .arg_specs
-                            .iter()
-                            .map(|arg_spec| {
-                                format!(
-                                    "{}{}{}",
-                                    TerseBool(arg_spec.batchable),
-                                    TerseBool(arg_spec.expandable),
-                                    TerseBool(arg_spec.ban_non_symbolic_size_expand)
-                                )
-                            })
-                            .collect::<Vec<_>>()
-                            .join(",")
-                    }
-                }
                 Circuit::Array(ac) => {
                     if self.sync_tensors {
                         ac.save_rrfs(true)?;
@@ -597,14 +739,6 @@ impl PrintOptions {
                 _ => "".to_owned(),
             }
         });
-        if let Some(threshold) = self.print_size_threshold {
-            if circ.info().numel() > BigUint::from(threshold) && !circ.is_array() {
-                result.push_str(&format!(
-                    " # {}",
-                    color(&oom_fmt(circ.info().numel()), clicolor!(Red))
-                ));
-            }
-        }
 
         for _ in 0..2 {
             result = result
@@ -618,7 +752,7 @@ impl PrintOptions {
     #[pyo3(name = "repr")]
     #[args(circuits = "*")]
     pub fn repr_circuits(&self, circuits: Vec<CircuitRc>) -> Result<String> {
-        let mut seen_hashes: HashMap<HashBytes, String> = HashMap::default();
+        let mut seen_hashes = HashMap::default();
 
         let disallow_name_ident = if !self.force_use_serial_numbers {
             let mut multiple_with_name: HashMap<String, bool> = HashMap::default();
@@ -647,6 +781,11 @@ impl PrintOptions {
                 for child in circ.children() {
                     recurse(child, seen, multiple_with_name)
                 }
+                if let Some(m) = circ.as_module() {
+                    for spec in &m.spec.arg_specs {
+                        recurse(spec.symbol.crc(), seen, multiple_with_name)
+                    }
+                }
             }
             for c in &circuits {
                 recurse(c.clone(), &mut seen, &mut multiple_with_name)
@@ -661,34 +800,24 @@ impl PrintOptions {
             circ: &Circuit,
             depth: usize,
             result: &mut String,
-            seen_hashes: &mut HashMap<HashBytes, String>,
+            seen_hashes: &mut HashMap<HashBytes, (String, String)>,
             runnning_serial_number: &mut usize,
+            runnning_leaf_number: &mut usize,
             selfy: &PrintOptions,
             is_last_child: &Vec<bool>,
             traversal: Option<OpaqueIterativeMatcherVal>,
             disallow_name_ident: Option<&HashMap<String, bool>>,
-            node_comments: Vec<String>,
+            parent_info: Option<(&Circuit, usize)>,
+            mut node_comments: Vec<String>,
         ) -> Result<()> {
             result.push_str(&last_child_arrows(is_last_child, true, selfy.arrows));
-            if let Some(prev) = seen_hashes.get(&circ.info().hash) {
-                result.push_str(prev);
-                for comment in node_comments {
-                    if !comment.is_empty() {
-                        write!(result, " # {}", comment).unwrap();
-                    }
-                }
-                result.push('\n');
-                return Ok(());
-            }
-            if let Some(ident) = selfy.reference_circuits.get(&circ.clone().rc()) {
-                result.push_str(&selfy.string_escape(ident));
-                result.push('\n');
-                return Ok(());
-            }
-
             let mut used_color = false;
             if let Some(colorer) = &selfy.colorer {
-                if let Some(color) = colorer.call(circ.clone().rc())?.0 {
+                if let Some(color) = colorer
+                    .call(circ.clone().rc())
+                    .context("colorer errored in print")?
+                    .0
+                {
                     used_color = true;
                     result.push_str(&format!(
                         "\u{001b}[{}m",
@@ -697,68 +826,75 @@ impl PrintOptions {
                 }
             }
 
-            // allow_name_ident is_some iff !self.force_use_serial_numbers
-            let disallow_name = disallow_name_ident
-                .zip(circ.name())
-                .map(|(map, name)| map[name])
-                .unwrap_or(true);
+            let term_on_repeat_or_reference = selfy.repr_full_line(
+                circ,
+                result,
+                seen_hashes,
+                runnning_serial_number,
+                disallow_name_ident,
+            )?;
 
-            let (repeat_string, line_prefix) = match (disallow_name, circ.name()) {
-                (_, Some(name)) => {
-                    let escaped = selfy.string_escape(name);
-                    let out = if disallow_name {
-                        let this_serial_number = runnning_serial_number.to_string();
-                        *runnning_serial_number += 1;
-                        this_serial_number + " " + &escaped
-                    } else {
-                        escaped
-                    };
-                    (out.clone(), out)
-                }
-                (true, None) => {
-                    let this_serial_number = runnning_serial_number.to_string();
-                    *runnning_serial_number += 1;
-                    (
-                        this_serial_number.clone() + " " + &circ.variant_string(),
-                        this_serial_number,
-                    )
-                }
-
-                (false, None) => unreachable!(),
-            };
-
-            seen_hashes.insert(circ.info().hash, repeat_string);
-
-            result.push_str(&line_prefix);
-            result.push_str(
-                &selfy
-                    .repr_line_info(circ.clone().rc())
-                    .with_context(|| format!("repr line info failed for circuit={:?}", circ))?,
-            );
             if used_color {
                 result.push_str("\u{001b}[0m");
             }
 
+            let info_from_parent = parent_info
+                .map(|(parent, child_idx)| {
+                    selfy.get_child_info(
+                        parent,
+                        seen_hashes,
+                        runnning_serial_number,
+                        disallow_name_ident,
+                        child_idx,
+                    )
+                })
+                .flatten();
+            if let Some(info) = info_from_parent {
+                result.push_str(&format!(" ! {}", info));
+            }
+
             let num_children = circ.num_children();
 
-            let new_traversal_per =
-                if let Some(new_traversal_per) = op_traversal_per(traversal, circ.crc())? {
-                    new_traversal_per
-                } else {
+            let new_traversal_per_op = if !term_on_repeat_or_reference {
+                let new_traversal_per_op = op_traversal_per(traversal, circ.crc())
+                    .context("traversal failed in printer")?;
+                if new_traversal_per_op.is_none() && circ.children().count() != 0 {
                     // all finished case
                     // traversal can't be used with bijection, so we can add ... without having to parse later
-                    if circ.children().count() != 0 {
-                        result.push_str(" ...");
-                    }
-                    result.push('\n');
-                    return Ok(());
-                };
+                    result.push_str(" ...");
+                }
+                new_traversal_per_op
+            } else {
+                None
+            };
+
+            let leaf_terming_here = circ.num_children() == 0 || new_traversal_per_op.is_none();
+
+            if selfy.number_leaves && leaf_terming_here {
+                node_comments.push(color(
+                    &format!("{}", *runnning_leaf_number),
+                    clicolor!(Magenta),
+                ));
+                *runnning_leaf_number += 1;
+            }
+
+            // TODO: move to commenter
+            if let Some(threshold) = selfy.print_size_threshold && circ.info().numel() > BigUint::from(threshold) && !circ.is_array() {
+                node_comments.push(color(&oom_fmt(circ.info().numel()), clicolor!(Red)));
+            }
+
             write_comment(result, node_comments, &selfy.commenters, circ.crc())?;
             result.push('\n');
-            if selfy.seen_children_same_line {
+
+            if term_on_repeat_or_reference || new_traversal_per_op.is_none() {
+                return Ok(());
+            }
+            let new_traversal_per = new_traversal_per_op.unwrap();
+
+            if selfy.seen_children_same_line && !selfy.has_child_info(circ) {
                 let child_prints_seen = circ
                     .children()
-                    .filter_map(|c| seen_hashes.get(&c.info().hash))
+                    .filter_map(|c| seen_hashes.get(&c.info().hash).as_ref().map(|x| &x.0))
                     .join(HORIZ_BAR);
                 let unseen: Vec<_> = circ
                     .children()
@@ -771,7 +907,7 @@ impl PrintOptions {
                         &is_last_child
                             .iter()
                             .copied()
-                            .chain(std::iter::once(unseen.len() == 0))
+                            .chain(iter::once(unseen.len() == 0))
                             .collect(),
                         true,
                         selfy.arrows,
@@ -784,7 +920,7 @@ impl PrintOptions {
                     let new_last_child = is_last_child
                         .iter()
                         .copied()
-                        .chain(std::iter::once(*i == unseen.len() - 1))
+                        .chain(iter::once(*i == unseen.len() - 1))
                         .collect();
                     recurse(
                         child,
@@ -792,10 +928,12 @@ impl PrintOptions {
                         result,
                         seen_hashes,
                         runnning_serial_number,
+                        runnning_leaf_number,
                         selfy,
                         &new_last_child,
                         new_traversal.clone(),
                         disallow_name_ident,
+                        None,
                         vec![],
                     )?;
                 }
@@ -814,10 +952,12 @@ impl PrintOptions {
                     result,
                     seen_hashes,
                     runnning_serial_number,
+                    runnning_leaf_number,
                     selfy,
                     &is_last_child,
                     new_traversal_per[0].clone(),
                     disallow_name_ident,
+                    Some((circ, 0)),
                     vec![],
                 )?;
                 return Ok(());
@@ -826,7 +966,7 @@ impl PrintOptions {
                 let new_last_child = is_last_child
                     .iter()
                     .copied()
-                    .chain(std::iter::once(i == num_children - 1))
+                    .chain(iter::once(i == num_children - 1))
                     .collect();
                 let child_specific_commenters = if selfy.comment_arg_names {
                     get_child_comments(circ, i)
@@ -839,10 +979,12 @@ impl PrintOptions {
                     result,
                     seen_hashes,
                     runnning_serial_number,
+                    runnning_leaf_number,
                     selfy,
                     &new_last_child,
                     new_traversal,
                     disallow_name_ident,
+                    Some((circ, i)),
                     child_specific_commenters,
                 )?;
             }
@@ -850,6 +992,7 @@ impl PrintOptions {
         }
         let mut result = String::new();
         let mut runnning_serial_number = 0;
+        let mut runnning_leaf_number = 0;
         for circuit in circuits {
             recurse(
                 &**circuit,
@@ -857,10 +1000,12 @@ impl PrintOptions {
                 &mut result,
                 &mut seen_hashes,
                 &mut runnning_serial_number,
+                &mut runnning_leaf_number,
                 &self,
                 &vec![],
                 self.traversal.clone(),
                 disallow_name_ident.as_ref(),
+                None,
                 vec![],
             )?;
         }
@@ -932,24 +1077,15 @@ impl PrintOptions {
     }
 
     #[staticmethod]
-    pub fn circuit_set_colorer(
-        map: HashMap<CliColor, HashSet<CircuitRc>>,
-    ) -> Result<CircuitToColorCode> {
+    pub fn circuit_set_colorer(map: HashMap<CliColor, HashSet<CircuitRc>>) -> CircuitToColorCode {
         let map: HashMap<HashBytes, CliColor> = map
             .into_iter()
-            .map(|(color, c)| {
-                Ok(c.into_iter()
-                    .map(|c| (c.info().hash, color))
-                    .collect::<Vec<_>>())
-            })
-            .collect::<Result<Vec<_>>>()?
-            .into_iter()
-            .flatten()
+            .flat_map(|(color, c)| c.into_iter().map(move |c| (c.info().hash, color)))
             .collect();
 
-        Ok(CircuitToColorCode::new_dyn(Box::new(move |c| {
+        CircuitToColorCode::new_dyn(Box::new(move |c| {
             Ok(map.get(&c.info().hash).cloned().into())
-        })))
+        }))
     }
 }
 
@@ -1034,7 +1170,7 @@ pub fn print_circuit_stats(circuit: &Circuit) {
     result.push_str(&format!(
         " nodes {} max_size {} flops {}",
         count_nodes(circuit.crc()),
-        oom_fmt(circuit.info().max_non_leaf_size.clone()),
+        oom_fmt(max_non_leaf_size(circuit.clone().rc())),
         oom_fmt(total_flops(circuit.crc()))
     ));
     println!("{}", result);
@@ -1100,6 +1236,16 @@ pub fn get_child_comments(parent: &Circuit, child_pos: usize) -> Vec<String> {
         Circuit::StoredCumulantVar(scv) => {
             format!("K{}", scv.cumulants.keys().nth(child_pos).unwrap())
         }
+        Circuit::Module(module) => {
+            if child_pos == 0 {
+                "Spec".to_owned()
+            } else if child_pos % 2 == 1 {
+                "".to_owned()
+            } else {
+                let spec_pos = (child_pos - 2) / 2;
+                "Arg ".to_owned() + module.spec.arg_specs[spec_pos].symbol.name().unwrap_or("")
+            }
+        }
         _ => "".to_owned(),
     }]
 }
@@ -1113,7 +1259,11 @@ pub fn write_comment(
     for comment in node_comments.into_iter().chain(
         commenters
             .iter()
-            .map(|commenter| commenter.call(circuit.clone()))
+            .map(|commenter| {
+                commenter
+                    .call(circuit.clone())
+                    .context("commenter errored in print")
+            })
             .collect::<Result<Vec<String>>>()?,
     ) {
         if !comment.is_empty() {
