@@ -558,6 +558,7 @@ pub fn concat_elim_identity(concat: &Concat) -> Option<CircuitRc> {
 }
 
 pub fn get_removable_axes(circuit: &CircuitRc) -> HashSet<usize> {
+    // plausibly this should have a flag for conforming to input shape
     match &***circuit {
         Circuit::Scalar(_scalar) => (0..circuit.info().rank()).collect(),
         Circuit::Rearrange(rearrange) => rearrange
@@ -571,6 +572,18 @@ pub fn get_removable_axes(circuit: &CircuitRc) -> HashSet<usize> {
 }
 
 pub fn remove_axes(circuit: &CircuitRc, axes: &HashSet<usize>) -> Option<CircuitRc> {
+    remove_axes_gen(circuit, axes, true)
+}
+
+pub fn remove_axes_weak(circuit: &CircuitRc, axes: &HashSet<usize>) -> Option<CircuitRc> {
+    remove_axes_gen(circuit, axes, false)
+}
+
+pub fn remove_axes_gen(
+    circuit: &CircuitRc,
+    axes: &HashSet<usize>,
+    rearrange_conform_for_remove_1: bool,
+) -> Option<CircuitRc> {
     if axes.is_empty() {
         return Some(circuit.clone());
     }
@@ -590,7 +603,11 @@ pub fn remove_axes(circuit: &CircuitRc, axes: &HashSet<usize>) -> Option<Circuit
         ),
         Circuit::Rearrange(rearrange) => {
             // canonicalize first to make input 1s (), allowing us to remove them from output
-            let spec = rearrange.conform_to_input_shape_spec().canonicalize(true);
+            let spec = if rearrange_conform_for_remove_1 {
+                rearrange.conform_to_input_shape_spec().canonicalize(true)
+            } else {
+                (*rearrange.spec).clone()
+            };
             Some(Rearrange::nrc_elim_identity(
                 rearrange.node.clone(),
                 spec.filter_out_axes(axes).unwrap(),
@@ -1556,7 +1573,7 @@ pub fn push_down_index_raw(
             } else {
                 bail!(PushDownIndexError::RearrangeNotPossible {
                     index_node: node.clone(),
-                    spec: inner.spec.clone()
+                    spec: (*inner.spec).clone()
                 })
             };
             let new_index = on_sub(inner.node.clone(), index, None)?;
