@@ -3,7 +3,7 @@ use std::io::prelude::*;
 use anyhow::{bail, Context, Result};
 use circuit_base::{
     flat_concat, get_compatible_dtype,
-    module::conform_all_modules,
+    module::{any_children_with_symbolic_sizes, conform_all_modules},
     print::{print_circuit_stats, PrintOptions},
     CircuitNode, CircuitRc, ModuleSpec,
 };
@@ -11,7 +11,7 @@ use num_bigint::BigUint;
 use pyo3::prelude::*;
 use rr_util::{
     opt_einsum::EinsumSpec,
-    py_types::Tensor,
+    py_types::{is_python_running, Tensor, PY_UTILS},
     python_println,
     repr::RustRepr,
     tensor_util::Shape,
@@ -27,6 +27,7 @@ use crate::{
     deep_rewrite::{
         compiler_simp, deep_heuristic_nest_adds, deep_optimize_einsums, deep_push_down_index_raw,
     },
+    module_rewrite::deep_module_remove_unused_inputs,
     scheduled_execution::{
         circuit_to_schedule, circuit_to_schedule_naive_toposort, Schedule, SchedulingError,
     },
@@ -173,6 +174,7 @@ impl OptimizationSettings {
 pub struct OptimizationCache {
     pub simplified: HashMap<HashBytes, CircuitRc>,
     pub distributed: HashMap<HashBytes, Option<CircuitRc>>,
+    pub module_removed_unused: HashMap<HashBytes, Option<CircuitRc>>,
     pub flops: HashMap<HashBytes, BigUint>,
     pub sum_of_node_sizes: HashMap<HashBytes, BigUint>,
     pub canonicalized: HashMap<HashBytes, Option<CircuitRc>>,
@@ -230,7 +232,10 @@ impl OptimizationContext {
 
 #[pyfunction(settings = "Default::default()")]
 #[pyo3(name = "optimize_circuit")]
-pub fn optimize_circuit_py(circuit: CircuitRc, settings: OptimizationSettings) -> CircuitRc {
+pub fn optimize_circuit_py(
+    circuit: CircuitRc,
+    settings: OptimizationSettings,
+) -> Result<CircuitRc> {
     let mut context = OptimizationContext {
         settings,
         ..Default::default()
@@ -238,11 +243,24 @@ pub fn optimize_circuit_py(circuit: CircuitRc, settings: OptimizationSettings) -
     optimize_circuit(circuit, &mut context)
 }
 
-pub fn optimize_circuit(circuit: CircuitRc, context: &mut OptimizationContext) -> CircuitRc {
+pub fn optimize_circuit(
+    circuit: CircuitRc,
+    context: &mut OptimizationContext,
+) -> Result<CircuitRc> {
     let print_timings = context.settings.verbose >= 2;
     let circuit = timed!(strip_names(circuit, context), 10, print_timings);
     let circuit = remove_autotags(circuit);
     let circuit = conform_all_modules(circuit);
+    if is_python_running() && any_children_with_symbolic_sizes(circuit.clone()) {
+        Python::with_gil(|py| {
+            pyo3::PyErr::warn(
+                py,
+                PY_UTILS.optimizing_symbolic_size_warning.as_ref(py),
+                "you're optimizing with symbolic sizes. This is not recommended!",
+                1,
+            )
+        })?;
+    }
     let circuit = timed!(deep_canonicalize(circuit, context), 10, print_timings);
     if context.settings.verbose >= 4 {
         python_println!("Original Circuit");
@@ -258,6 +276,15 @@ pub fn optimize_circuit(circuit: CircuitRc, context: &mut OptimizationContext) -
         file.write_all(s.as_bytes()).unwrap();
     }
     let circuit = timed!(compiler_simp(circuit, context), 10, print_timings);
+    let circuit = timed!(
+        apply_fn_until_same(&circuit, |x| deep_module_remove_unused_inputs(
+            compiler_simp(x.clone(), context),
+            &mut context.cache.module_removed_unused,
+            true
+        )),
+        10,
+        print_timings
+    );
     // originally tried push_down_index in compiler_simp, but that produced worse circuits
     // for unknown reasons, maybe i'll investigate
     let circuit = timed!(
@@ -307,7 +334,7 @@ pub fn optimize_circuit(circuit: CircuitRc, context: &mut OptimizationContext) -
             .unwrap();
         print_circuit_stats(&*circuit);
     }
-    circuit
+    Ok(circuit)
 }
 
 #[pyfunction(settings = "Default::default()")]
@@ -325,7 +352,7 @@ pub fn optimize_to_schedule(
             .print(circuit.clone())
             .context("capture print fail!")?;
     }
-    let optimized_circuit = timed!(optimize_circuit(circuit, context), 10, verbose >= 1);
+    let optimized_circuit = timed!(optimize_circuit(circuit, context), 10, verbose >= 1)?;
     // return evaluate_fn(&*optimized_circuit).unwrap();
     let schedule = if settings.scheduling_naive {
         circuit_to_schedule_naive_toposort(optimized_circuit)
@@ -344,12 +371,10 @@ pub fn optimize_to_schedule(
 
 #[pyfunction(settings = "Default::default()")]
 pub fn optimize_and_evaluate(circuit: CircuitRc, settings: OptimizationSettings) -> Result<Tensor> {
-    if !circuit.info().is_explicitly_computable {
-        bail!(SchedulingError::NotExplicitlyComputable { circuit });
-    }
-    let schedule = optimize_to_schedule(circuit, settings)?;
-
-    Ok(schedule.evaluate(settings))
+    Ok(optimize_and_evaluate_many(vec![circuit], settings)?
+        .into_iter()
+        .next()
+        .unwrap())
 }
 
 /// in python, lots of functions take in collections of circuits and operate on them at once
@@ -363,6 +388,12 @@ pub fn optimize_and_evaluate_many(
     circuits: Vec<CircuitRc>,
     settings: OptimizationSettings,
 ) -> Result<Vec<Tensor>> {
+    for c in &circuits {
+        if !c.info().is_explicitly_computable {
+            bail!(SchedulingError::NotExplicitlyComputable { circuit: c.clone() });
+        }
+    }
+
     let schedule = optimize_to_schedule_many(circuits, settings)?;
     Ok(schedule.evaluate_many(settings))
 }
