@@ -2,9 +2,10 @@ use std::fmt::{self, Display};
 
 use anyhow::{bail, Context, Result};
 use circuit_base::{
+    cached_circuit_properties::max_non_leaf_size,
     circuit_utils::{hash_to_node_non_free, toposort_circuit},
     get_compatible_dtype,
-    module::substitute_all_modules,
+    module::{any_children_with_symbolic_sizes, substitute_all_modules},
     prelude::*,
     print::{oom_fmt, PrintOptions},
     visit_circuit_non_free, Add, Array, IrreducibleNode, Leaf, ModuleSpec, Scalar, Symbol,
@@ -16,9 +17,12 @@ use num_bigint::BigUint;
 use pyo3::{exceptions::PyValueError, prelude::*, types::PyBytes};
 use rr_util::{
     lru_cache::TensorCacheRrfs,
-    py_types::{scalar_to_tensor, tensor_scale, un_flat_concat, ExtraPySelfOps, Tensor, PY_UTILS},
+    py_types::{
+        is_python_running, scalar_to_tensor, tensor_scale, un_flat_concat, ExtraPySelfOps, Tensor,
+        PY_UTILS,
+    },
     pycall, python_error_exception, sv,
-    tensor_util::{Shape, TorchDeviceDtype},
+    tensor_util::{Shape, TorchDevice, TorchDeviceDtype},
     timed,
     util::{arc_ref_clone, with_context_failable, HashBytes},
 };
@@ -110,7 +114,7 @@ impl Schedule {
             } else {
                 evaluate_schedule(self)[&self.output_circuit.as_ref().unwrap().0].clone()
             };
-            if self.device_dtype.device != *"cpu" {
+            if self.device_dtype.device != TorchDevice::Cpu {
                 Python::with_gil(|py| {
                     PY_UTILS
                         .torch
@@ -192,7 +196,7 @@ impl Schedule {
     #[pyo3(name = "deserialize")]
     pub fn deserialize_py(
         string: String,
-        device: String,
+        device: TorchDevice,
         tensor_cache: Option<TensorCacheRrfs>,
     ) -> Result<Self> {
         let mut tensor_cache = tensor_cache;
@@ -219,7 +223,7 @@ impl Schedule {
 impl Schedule {
     pub fn deserialize(
         string: String,
-        device: String,
+        device: TorchDevice,
         tensor_cache: &mut Option<TensorCacheRrfs>,
     ) -> Result<Self> {
         let sent: ScheduleToSend =
@@ -684,10 +688,25 @@ pub fn circuit_to_schedule(
     circuit: CircuitRc,
     context: &mut OptimizationContext,
 ) -> Result<Schedule> {
+    if is_python_running() && any_children_with_symbolic_sizes(circuit.clone()) {
+        Python::with_gil(|py| {
+            pyo3::PyErr::warn(
+                py,
+                PY_UTILS.optimizing_symbolic_size_warning.as_ref(py),
+                concat!(
+                    "you're scheduling with symbolic sizes.",
+                    " Perhaps you need to conform_all_modules?",
+                    " (this is typically done in optimize_circuit)."
+                ),
+                1,
+            )
+        })?;
+    }
+
     let mut dag = circuits_to_dag(&[circuit.clone()]);
-    if circuit.info().max_non_leaf_size > BigUint::from(context.cache.fallback_total_numel) {
+    if max_non_leaf_size(circuit.clone()) > BigUint::from(context.cache.fallback_total_numel) {
         bail!(SchedulingOOMError::Single {
-            numel: circuit.info().max_non_leaf_size.clone(),
+            numel: max_non_leaf_size(circuit.clone()),
         });
     }
     let order_result = {
