@@ -1,16 +1,17 @@
 use std::{
-    collections::BTreeSet,
+    collections::{BTreeMap, BTreeSet},
     fmt::{self, Debug},
     ops,
     sync::Arc,
     vec,
 };
 
-use anyhow::{bail, Result};
+use anyhow::{bail, Context, Result};
 use circuit_base::{
     opaque_iterative_matcher::{
         Update as BaseUpdate, UpdatedIterativeMatcher as BaseUpdatedIterativeMatcher,
     },
+    set_of_circuits::SetOfCircuitIdentities,
     CircuitNode, CircuitRc,
 };
 use macro_rules_attribute::apply;
@@ -691,8 +692,7 @@ impl IterativeMatcher {
                         .enumerate()
                         .map(|(i, x)| {
                             x.and_then(|x| {
-                                (i != 0 && (i - 1) % 2 != 0)
-                                    .then(|| IterativeMatcherData::NoModuleSpec(x).into())
+                                (i != 0).then(|| IterativeMatcherData::NoModuleSpec(x).into())
                             })
                             .into()
                         })
@@ -913,9 +913,15 @@ impl IterativeMatcher {
                 }
             }
             IterativeMatcherData::Raw(f) => f.0(circuit)?,
-            IterativeMatcherData::PyFunc(pyfunc) => {
-                Python::with_gil(|py| pyfunc.call1(py, (circuit,)).and_then(|r| r.extract(py)))?
-            }
+            IterativeMatcherData::PyFunc(pyfunc) => Python::with_gil(|py| {
+                pyfunc
+                    .call1(py, (circuit,))
+                    .context("calling python iterative matcher failed")
+                    .and_then(|r| {
+                        r.extract(py)
+                            .context("extracting from python iterative matcher failed")
+                    })
+            })?,
         };
 
         if let Some(UpdatedIterativeMatcher::Many(items)) = &res.updated {
@@ -1172,6 +1178,19 @@ impl IterativeMatcher {
         Getter::default().get_unique(circuit, self_.crc(), Some(fancy_validate))
     }
 
+    pub fn get_paths(&self, circuit: CircuitRc) -> Result<BTreeMap<CircuitRc, Vec<CircuitRc>>> {
+        Getter::default().get_paths(circuit, self_.crc())
+    }
+    pub fn get_all_paths(
+        &self,
+        circuit: CircuitRc,
+    ) -> Result<BTreeMap<CircuitRc, Vec<Vec<CircuitRc>>>> {
+        Getter::default().get_all_paths(circuit, self_.crc())
+    }
+    pub fn get_all_circuits_in_paths(&self, circuit: CircuitRc) -> Result<SetOfCircuitIdentities> {
+        Getter::default().get_all_circuits_in_paths(circuit, self_.crc())
+    }
+
     pub fn validate(&self, circuit: CircuitRc) -> Result<()> {
         Getter::default().validate(circuit, self_.crc())
     }
