@@ -1,13 +1,22 @@
-use std::collections::{BTreeMap, BTreeSet};
+use std::{
+    collections::{BTreeMap, BTreeSet},
+    sync::Arc,
+};
 
 use anyhow::{anyhow, bail, Result};
-use circuit_base::{expand_node::expand_node, CircuitNode, CircuitRc};
+use circuit_base::{
+    expand_node::{expand_node, ReplaceMapRc},
+    set_of_circuits::SetOfCircuitIdentities,
+    CircuitNode, CircuitRc,
+};
 use macro_rules_attribute::apply;
 use pyo3::prelude::*;
 use rr_util::{
     cached_method,
+    eq_by_big_hash::EqByBigHash,
     util::{transpose, HashBytes},
 };
+use rustc_hash::FxHashSet as HashSet;
 
 use crate::{
     iterative_matcher::{all_finished, function_per_child, per_child, per_child_with_term},
@@ -176,6 +185,10 @@ pub struct Getter {
     #[pyo3(get, set)]
     pub(super) default_fancy_validate: bool,
     pub(super) cache: cached::UnboundCache<(HashBytes, IterativeMatcherRc), BTreeSet<CircuitRc>>,
+    pub(super) paths_cache:
+        cached::UnboundCache<(HashBytes, IterativeMatcherRc), BTreeMap<CircuitRc, Vec<CircuitRc>>>,
+    pub(super) all_paths_cache:
+        cached::UnboundCache<(HashBytes, IterativeMatcherRc), Vec<CircuitLinkedList>>,
 }
 
 impl Default for Getter {
@@ -183,10 +196,37 @@ impl Default for Getter {
         Self {
             default_fancy_validate: false,
             cache: cached::UnboundCache::new(),
+            paths_cache: cached::UnboundCache::new(),
+            all_paths_cache: cached::UnboundCache::new(),
         }
     }
 }
 
+#[derive(Clone, Debug)]
+pub enum CircuitLinkedList {
+    None,
+    Some(CircuitRc, Arc<CircuitLinkedList>),
+    // this is arc bc its used in cached crate which uses mutex blah blah blah
+    // it isn't actually meant to (or likely actually) go between threads
+    // maybe i should just unsafe impl send instead
+}
+
+impl CircuitLinkedList {
+    fn list(&self) -> Vec<CircuitRc> {
+        let mut out = vec![];
+        let mut head = self;
+        loop {
+            if let CircuitLinkedList::Some(c, n) = &head {
+                out.push(c.clone());
+                head = &*n;
+            } else {
+                break;
+            }
+        }
+        out
+    }
+}
+
 #[pymethods]
 impl Getter {
     #[new]
@@ -221,6 +261,37 @@ impl Getter {
         Ok(out)
     }
 
+    pub fn get_paths(
+        &mut self,
+        circuit: CircuitRc,
+        matcher: IterativeMatcherRc,
+    ) -> Result<BTreeMap<CircuitRc, Vec<CircuitRc>>> {
+        self.get_paths_impl(circuit, matcher) // TODO: add validation?
+    }
+
+    pub fn get_all_circuits_in_paths(
+        &mut self,
+        circuit: CircuitRc,
+        matcher: IterativeMatcherRc,
+    ) -> Result<SetOfCircuitIdentities> {
+        Ok(self
+            .get_all_paths_impl(circuit, matcher)?
+            .into_values()
+            .flatten()
+            .flatten()
+            .map(|c| c.info().hash)
+            .collect::<HashSet<HashBytes>>()
+            .into()) // TODO: add validation?
+    }
+
+    pub fn get_all_paths(
+        &mut self,
+        circuit: CircuitRc,
+        matcher: IterativeMatcherRc,
+    ) -> Result<BTreeMap<CircuitRc, Vec<Vec<CircuitRc>>>> {
+        self.get_all_paths_impl(circuit, matcher) // TODO: add validation?
+    }
+
     pub fn get_unique_op(
         &mut self,
         circuit: CircuitRc,
@@ -286,6 +357,92 @@ impl Getter {
         }
         Ok(out)
     }
+
+    fn get_all_paths_impl(
+        &mut self,
+        circuit: CircuitRc,
+        matcher: IterativeMatcherRc,
+    ) -> Result<BTreeMap<CircuitRc, Vec<Vec<CircuitRc>>>> {
+        let result_raw = self.get_all_paths_impl_intermediate(circuit, matcher)?;
+        let mut result: BTreeMap<CircuitRc, Vec<Vec<CircuitRc>>> = Default::default();
+        for ll in result_raw {
+            let mut listy = ll.list();
+            listy.reverse();
+            let foundy = listy[0].clone();
+            if let Some(existing) = result.get_mut(&foundy) {
+                existing.push(listy);
+            } else {
+                result.insert(foundy, vec![listy]);
+            }
+        }
+        Ok(result)
+    }
+
+    #[apply(cached_method)]
+    #[self_id(self_)]
+    #[key((circuit.info().hash, matcher.clone()))]
+    #[use_try]
+    #[cache_expr(all_paths_cache)]
+    fn get_all_paths_impl_intermediate(
+        &mut self,
+        circuit: CircuitRc,
+        matcher: IterativeMatcherRc,
+    ) -> Result<Vec<CircuitLinkedList>> {
+        let IterateMatchResults { updated, found } = matcher.match_iterate(circuit.clone())?;
+
+        let mut out: Vec<CircuitLinkedList> = Default::default();
+        if found {
+            out.push(CircuitLinkedList::Some(
+                circuit.clone(),
+                Arc::new(CircuitLinkedList::None),
+            ));
+        }
+        if !all_finished(&updated) {
+            let new_matchers = per_child(updated, matcher, circuit.num_children());
+            for (child, new_matcher) in circuit.children().zip(new_matchers) {
+                if let Some(new_matcher) = new_matcher {
+                    let child_paths = self_.get_all_paths_impl_intermediate(child, new_matcher)?;
+                    out.extend(
+                        child_paths
+                            .into_iter()
+                            .map(|ll| CircuitLinkedList::Some(circuit.clone(), Arc::new(ll))),
+                    );
+                }
+            }
+        }
+        Ok(out)
+    }
+
+    #[apply(cached_method)]
+    #[self_id(self_)]
+    #[key((circuit.info().hash, matcher.clone()))]
+    #[use_try]
+    #[cache_expr(paths_cache)]
+    fn get_paths_impl(
+        &mut self,
+        circuit: CircuitRc,
+        matcher: IterativeMatcherRc,
+    ) -> Result<BTreeMap<CircuitRc, Vec<CircuitRc>>> {
+        let IterateMatchResults { updated, found } = matcher.match_iterate(circuit.clone())?;
+
+        let mut out: BTreeMap<CircuitRc, Vec<CircuitRc>> = Default::default();
+        if found {
+            out.insert(circuit.clone(), vec![circuit.clone()]);
+        }
+        if !all_finished(&updated) {
+            let new_matchers = per_child(updated, matcher, circuit.num_children());
+            for (child, new_matcher) in circuit.children().zip(new_matchers) {
+                if let Some(new_matcher) = new_matcher {
+                    let child_paths = self_.get_paths_impl(child, new_matcher)?;
+                    for (child, mut path) in child_paths.into_iter() {
+                        path.push(circuit.clone());
+                        out.insert(child, path);
+                    }
+                }
+            }
+        }
+        Ok(out)
+    }
 }
 
 #[pyclass]
@@ -343,6 +500,17 @@ impl BoundGetter {
             .get_unique(circuit, self.matcher.clone(), fancy_validate)
     }
 
+    pub fn get_paths(&mut self, circuit: CircuitRc) -> Result<BTreeMap<CircuitRc, Vec<CircuitRc>>> {
+        self.getter.get_paths(circuit, self.matcher.clone())
+    }
+
+    pub fn get_all_paths(
+        &mut self,
+        circuit: CircuitRc,
+    ) -> Result<BTreeMap<CircuitRc, Vec<Vec<CircuitRc>>>> {
+        self.getter.get_all_paths(circuit, self.matcher.clone())
+    }
+
     pub fn validate(&mut self, circuit: CircuitRc) -> Result<()> {
         self.getter.validate(circuit, self.matcher.clone())
     }
@@ -354,6 +522,12 @@ pub struct AnyFound {
     pub(super) cache: cached::UnboundCache<(HashBytes, IterativeMatcherRc), bool>,
 }
 
+impl Default for AnyFound {
+    fn default() -> Self {
+        Self::new()
+    }
+}
+
 #[pymethods]
 impl AnyFound {
     #[new]
@@ -462,14 +636,8 @@ pub struct Expander {
     pub default_fancy_validate: bool,
     #[pyo3(get)]
     pub(super) suffix: Option<String>,
-    pub(super) batch_cache: cached::UnboundCache<
-        (
-            HashBytes,
-            Vec<IterativeMatcherRc>,
-            BTreeMap<CircuitRc, CircuitRc>,
-        ),
-        CircuitRc,
-    >,
+    pub(super) batch_cache:
+        cached::UnboundCache<(HashBytes, Vec<IterativeMatcherRc>, HashBytes), CircuitRc>,
     pub(super) validation_getter: Getter,
 }
 
@@ -539,14 +707,14 @@ impl Expander {
 impl Expander {
     #[apply(cached_method)]
     #[self_id(self_)]
-    #[key((circuit.info().hash, matchers.clone(), extra_replacements.clone()))]
+    #[key((circuit.info().hash, matchers.clone(), extra_replacements.hash()))]
     #[use_try]
     #[cache_expr(batch_cache)]
     fn batch_impl(
         &mut self,
         circuit: CircuitRc,
         matchers: Vec<IterativeMatcherRc>,
-        extra_replacements: &BTreeMap<CircuitRc, CircuitRc>,
+        extra_replacements: &ReplaceMapRc,
     ) -> Result<CircuitRc> {
         let results = matchers
             .iter()
@@ -569,7 +737,7 @@ impl Expander {
             return Ok(replaced.clone());
         }
 
-        if results.iter().all(|x| all_finished(&x.updated)) {
+        if results.iter().all(|x| all_finished(&x.updated)) && extra_replacements.is_empty() {
             return Ok(circuit);
         }
 
@@ -589,18 +757,26 @@ impl Expander {
             .map(|(c, new_matchers)| self_.batch_impl(c, new_matchers, extra_replacements))
             .collect::<Result<_>>()?;
 
-        Ok(expand_node(circuit, &new_children, |c, rep, child_idx| {
+        let expanded = expand_node(circuit.clone(), &new_children, |c, rep, child_idx| {
             self_.batch_impl(
                 c,
                 new_matchers_per[child_idx].clone(),
-                &extra_replacements
-                    .iter()
-                    .map(|(a, b)| (a.clone(), b.clone()))
-                    .chain(rep)
-                    .collect(),
+                &ReplaceMapRc::new(
+                    extra_replacements
+                        .iter()
+                        .map(|(a, b)| (a.clone(), b.clone()))
+                        .chain(rep)
+                        .collect(),
+                ),
             )
-        })?
-        .add_suffix(self_.suffix.as_ref().map(|x| &**x)))
+        })?;
+        let expanded = if expanded != circuit {
+            expanded.add_suffix(self_.suffix.as_ref().map(|x| &**x))
+        } else {
+            expanded
+        };
+
+        Ok(expanded)
     }
 
     pub fn suffix(&self) -> Option<&str> {
