@@ -9,8 +9,8 @@ use anyhow::{bail, Context, Error, Result};
 use cached::Cached;
 use circuit_base::{
     cumulant::{dim_permutation_for_circuits, partitions},
-    visit_circuit, Add, CircResult, CircuitNode, CircuitRc, CircuitType, Cumulant, DiscreteVar,
-    Einsum, GeneralFunction, Index, Scalar,
+    visit_circuit, Add, Array, CircResult, CircuitNode, CircuitRc, CircuitType, Cumulant,
+    DiscreteVar, Einsum, GeneralFunction, Index, Scalar,
 };
 use circuit_rewrites::circuit_optimizer::{optimize_and_evaluate, OptimizationSettings};
 use itertools::Itertools;
@@ -18,9 +18,11 @@ use macro_rules_attribute::apply;
 use pyo3::{exceptions::PyValueError, prelude::*};
 use rr_util::{
     cached_method, fn_struct,
-    py_types::{PyShape, Tensor, PY_UTILS},
-    python_error_exception,
-    tensor_util::{Shape, Slice, TensorAxisIndex, TensorIndex},
+    py_types::{scalar_to_tensor, PyShape, Tensor, PY_UTILS},
+    python_error_exception, sv,
+    tensor_util::{
+        Shape, Slice, TensorAxisIndex, TensorIndex, TorchDeviceDtype, TorchDeviceDtypeOp,
+    },
     util::{flip_op_result, HashBytes},
 };
 use smallvec::SmallVec as Sv;
@@ -51,7 +53,7 @@ pub fn default_var_matcher() -> IterativeMatcherRc {
 
 pub trait SampleSpecTrait {
     /// Like the python method, but modify inplace (via mutex or other interior mutability)
-    fn sample_var(&self, d: &DiscreteVar) -> Result<CircuitRc>;
+    fn sample_var(&self, d: &DiscreteVar, device_dtype: TorchDeviceDtypeOp) -> Result<CircuitRc>;
     fn get_empirical_weights(&self) -> CircuitRc;
     fn get_overall_weight(&self) -> CircuitRc {
         Scalar::nrc(
@@ -61,23 +63,27 @@ pub trait SampleSpecTrait {
         )
     }
     fn get_sample_shape(&self) -> Shape;
-    fn get_transform(self) -> TransformRc
+    fn get_transform(self, device_dtype: TorchDeviceDtypeOp) -> TransformRc
     where
         Self: Sized + Sync + Send + 'static,
     {
-        Transform::new_func_err(move |circ| self.sample_var(circ.as_discrete_var_unwrap())).rc()
+        Transform::new_func_err(move |circ| {
+            self.sample_var(circ.as_discrete_var_unwrap(), device_dtype.clone())
+        })
+        .rc()
     }
     fn get_sample_expander(
         self,
         var_matcher: IterativeMatcherRc,
         default_fancy_validate: bool,
         suffix: Option<String>,
+        device_dtype: TorchDeviceDtypeOp,
     ) -> Expander
     where
         Self: Sized + Sync + Send + 'static,
     {
         Expander::new(
-            vec![(var_matcher, self.get_transform())],
+            vec![(var_matcher, self.get_transform(device_dtype))],
             true,
             default_fancy_validate,
             suffix,
@@ -109,8 +115,13 @@ macro_rules! py_forward_sample_spec {
         #[pymethods]
         impl $type {
             #[pyo3(name = "sample_var")]
-            fn sample_var_py(&self, d: &DiscreteVar) -> Result<CircuitRc> {
-                self.sample_var(d)
+            #[args(device_dtype = "TorchDeviceDtypeOp::NONE")]
+            fn sample_var_py(
+                &self,
+                d: &DiscreteVar,
+                device_dtype: TorchDeviceDtypeOp,
+            ) -> Result<CircuitRc> {
+                self.sample_var(d, device_dtype)
             }
             #[pyo3(name = "get_empirical_weights")]
             fn get_empirical_weights_py(&self) -> CircuitRc {
@@ -125,23 +136,30 @@ macro_rules! py_forward_sample_spec {
                 PyShape(self.get_sample_shape())
             }
             #[pyo3(name = "get_transform")]
-            fn get_transform_py(&self) -> TransformRc {
-                self.clone().get_transform()
+            #[args(device_dtype = "TorchDeviceDtypeOp::NONE")]
+            fn get_transform_py(&self, device_dtype: TorchDeviceDtypeOp) -> TransformRc {
+                self.clone().get_transform(device_dtype)
             }
             #[pyo3(name = "get_sample_expander")]
             #[args(
                 var_matcher = "default_var_matcher()",
                 default_fancy_validate = "false",
-                suffix = "\"sample\".to_owned()"
+                suffix = "\"sample\".to_owned()",
+                device_dtype = "TorchDeviceDtypeOp::NONE"
             )]
             fn get_sample_expander_py(
                 &self,
                 var_matcher: IterativeMatcherRc,
                 default_fancy_validate: bool,
                 suffix: Option<String>,
+                device_dtype: TorchDeviceDtypeOp,
             ) -> Expander {
-                self.clone()
-                    .get_sample_expander(var_matcher, default_fancy_validate, suffix)
+                self.clone().get_sample_expander(
+                    var_matcher,
+                    default_fancy_validate,
+                    suffix,
+                    device_dtype,
+                )
             }
         }
 
@@ -196,14 +214,28 @@ impl RandomSampleSpec {
 }
 
 impl RandomSampleSpec {
-    fn sample_impl(&self, probs_and_group: CircuitRc) -> Result<Tensor> {
+    fn sample_impl(
+        &self,
+        probs_and_group: CircuitRc,
+        device_dtype: TorchDeviceDtype,
+    ) -> Result<Tensor> {
         let key = probs_and_group.info().hash;
         if let Some(out) = self.indices_cache.lock().unwrap().cache_get(&key) {
             return Ok(out.clone());
         }
 
         let prob_tensor = optimize_and_evaluate(
-            probs_and_group.clone(),
+            Add::nrc(
+                vec![
+                    probs_and_group.clone(),
+                    Array::nrc(
+                        scalar_to_tensor(0., sv![], device_dtype)
+                            .context("failed to create scalar tensor???")?,
+                        None,
+                    ),
+                ],
+                None,
+            ),
             self.probs_and_group_evaluation_settings,
         )
         .context("failed to evaluate prob group")?;
@@ -238,6 +270,26 @@ impl Default for RandomSampleSpec {
     }
 }
 
+fn combine_dtype_device_for_sample(
+    var: &DiscreteVar,
+    device_dtype: TorchDeviceDtypeOp,
+) -> Result<TorchDeviceDtype> {
+    device_dtype
+        .clone()
+        .combine(var.info().device_dtype.clone())
+        .with_context(|| {
+            format!(
+                concat!(
+                    "sampled var device_dtype failed to combine with passed device_dtype",
+                    "var device_dtype={:?} device_dtype={:?}"
+                ),
+                &var.info().device_dtype,
+                device_dtype
+            )
+        })
+        .map(|x| x.unwrap_or_defaults())
+}
+
 fn sample_indices(
     input: Tensor,
     shape: Shape,
@@ -254,9 +306,10 @@ fn sample_indices(
     Ok(indices)
 }
 impl SampleSpecTrait for RandomSampleSpec {
-    fn sample_var(&self, var: &DiscreteVar) -> Result<CircuitRc> {
+    fn sample_var(&self, var: &DiscreteVar, device_dtype: TorchDeviceDtypeOp) -> Result<CircuitRc> {
+        let device_dtype = combine_dtype_device_for_sample(var, device_dtype)?;
         let indices = self
-            .sample_impl(var.probs_and_group.clone())?
+            .sample_impl(var.probs_and_group.clone(), device_dtype)?
             .flatten() // flatten here, unflatten via rearrange
             .try_into()
             .unwrap();
@@ -378,7 +431,7 @@ impl RunDiscreteVarAllSpec {
 }
 
 impl SampleSpecTrait for RunDiscreteVarAllSpec {
-    fn sample_var(&self, var: &DiscreteVar) -> CircResult {
+    fn sample_var(&self, var: &DiscreteVar, _device_dtype: TorchDeviceDtypeOp) -> CircResult {
         let this_group_i = match self
             .groups
             .iter()
@@ -478,8 +531,8 @@ impl SampleSpec {
 }
 
 impl SampleSpecTrait for SampleSpec {
-    fn sample_var(&self, d: &DiscreteVar) -> CircResult {
-        self.as_trait_obj().sample_var(d)
+    fn sample_var(&self, d: &DiscreteVar, device_dtype: TorchDeviceDtypeOp) -> CircResult {
+        self.as_trait_obj().sample_var(d, device_dtype)
     }
     fn get_empirical_weights(&self) -> CircuitRc {
         self.as_trait_obj().get_empirical_weights()
@@ -502,6 +555,8 @@ pub struct Sampler {
     cumulant_matcher: IterativeMatcherRc,
     #[pyo3(get)]
     sample_spec: SampleSpec,
+    #[pyo3(get)]
+    run_on_sampled: TransformRc,
     estimated_cache: cached::UnboundCache<HashBytes, CircuitRc>,
 }
 
@@ -513,22 +568,28 @@ impl Sampler {
     #[args(
         var_matcher = "default_var_matcher()",
         cumulant_matcher = "Matcher::types(vec![CircuitType::Cumulant]).into()",
-        suffix = "\"sample\".to_owned()"
+        suffix = "\"sample\".to_owned()",
+        device_dtype = "TorchDeviceDtypeOp::NONE",
+        run_on_sampled = "Transform::ident().rc()"
     )]
     fn new(
         sample_spec: SampleSpec,
         var_matcher: IterativeMatcherRc,
         cumulant_matcher: IterativeMatcherRc,
         suffix: Option<String>,
+        device_dtype: TorchDeviceDtypeOp,
+        run_on_sampled: TransformRc,
     ) -> Self {
-        let expander = sample_spec
-            .clone()
-            .get_sample_expander(var_matcher, false, suffix);
+        let expander =
+            sample_spec
+                .clone()
+                .get_sample_expander(var_matcher, false, suffix, device_dtype);
         Self {
             expander,
             cumulant_matcher,
             sample_spec,
             estimated_cache: cached::UnboundCache::new(),
+            run_on_sampled,
         }
     }
 
@@ -550,18 +611,24 @@ impl Sampler {
                 .chain(circuit.shape())
                 .cloned()
                 .collect::<Shape>();
-        assert!(is_orig_shape || is_sampled_shape);
-        if is_orig_shape {
-            Ok(out
-                .expand_at_axes(
-                    (0..self.sample_spec.get_sample_shape().len()).collect(),
-                    self.sample_spec.get_sample_shape().into_vec(),
-                    None,
-                )?
-                .rc())
+        assert!(
+            is_orig_shape || is_sampled_shape,
+            "out = {:?}, circuit = {:?}, sample = {:?}",
+            out.shape(),
+            circuit.shape(),
+            &self.sample_spec.get_sample_shape()
+        );
+        let out = if is_orig_shape {
+            out.expand_at_axes(
+                (0..self.sample_spec.get_sample_shape().len()).collect(),
+                self.sample_spec.get_sample_shape().into_vec(),
+                None,
+            )?
+            .rc()
         } else {
-            Ok(out)
-        }
+            out
+        };
+        self.run_on_sampled.run(out)
     }
     pub fn estimate_and_sample(&mut self, circuit: CircuitRc) -> Result<CircuitRc> {
         let circuit = self.estimate(circuit)?;
