@@ -34,26 +34,26 @@ pub fn push_down_index(
     elim_identity: bool,
 ) -> Result<CircuitRc> {
     #[apply(cached_lambda)]
-    #[key((index.info().hash, matcher.clone()), (HashBytes, IterativeMatcherRc))]
+    #[key((index.info().hash, traversal.clone()), (HashBytes, IterativeMatcherRc))]
     #[use_try]
-    fn push_down_rec(index: Index, matcher: IterativeMatcherRc) -> Result<CircuitRc> {
+    fn push_down_rec(index: Index, traversal: IterativeMatcherRc) -> Result<CircuitRc> {
         if elim_identity {
             if let Some(removed_node) = index_elim_identity(&index) {
                 return Ok(removed_node);
             }
         }
-        let updated = matcher
+        let updated = traversal
             .match_iterate(index.node.clone())?
-            .unwrap_or_same(matcher)
+            .unwrap_or_same(traversal)
             .0;
-        if let Some(new_matcher) = require_single(updated)
-            .context("push down index doesn't support matcher per child")?
+        if let Some(new_traversal) = require_single(updated)
+            .context("push down index doesn't support traversal per child")?
             .0
         {
             push_down_index_raw(
                 &index,
                 allow_partial_pushdown,
-                |new_index| push_down_rec(new_index, new_matcher.clone()),
+                |new_index| push_down_rec(new_index, new_traversal.clone()),
                 suffix.clone(),
             )
         } else {
