@@ -2,9 +2,7 @@ use std::vec;
 
 use anyhow::{bail, Context, Result};
 use circuit_base::{computational_node::EinsumArgs, Add, Circuit, CircuitNode, CircuitRc, Einsum};
-use circuit_rewrites::algebraic_rewrite::{
-    add_flatten_once, make_einsum_ints_same_one_layer_and_int_info,
-};
+use circuit_rewrites::algebraic_rewrite::make_einsum_ints_same_one_layer_and_int_info;
 use get_update_node::{
     iterative_matcher::UpdatedIterativeMatcher, IterativeMatcher, IterativeMatcherRc,
 };
@@ -472,7 +470,20 @@ impl NestFullInfo for NestAddsFullInfo {
         rec_on_children: &[FlattenImplRec<Self>],
         add: Add,
     ) -> FlattenImplRec<Self> {
-        let out_circ = add_flatten_once(&add).unwrap_or(add).rc();
+        let new_operands: Vec<CircuitRc> = add
+            .nodes
+            .iter()
+            .zip(rec_on_children)
+            .flat_map(|(node, (_, named_idxs, _))| {
+                if named_idxs.is_some() {
+                    node.as_add().unwrap().nodes.clone()
+                } else {
+                    vec![node.clone()]
+                }
+            })
+            .collect();
+        let out_circ = Add::nrc(new_operands, add.name_cloned());
+
         finish_with_extra_and_circ(&rec_on_children, &vec![(); rec_on_children.len()], out_circ)
     }
 
@@ -585,7 +596,7 @@ impl<I: NestFullInfo> NestIdxsItem<I> {
 }
 
 /// maintains invariant that indices are in sorted enumeration order
-#[derive(Clone)]
+#[derive(Clone, Debug)]
 pub struct NestEnumerationItem<I: NestFullInfo>(NestIdxsItem<I>);
 
 impl<I: NestFullInfo> NestEnumerationItem<I> {
