@@ -33,6 +33,7 @@ pub struct PyUtils {
     pub tensor_from_bytes: PyObject,
     pub assert_tensors_close: PyObject,
     pub builtins: Py<PyModule>,
+    pub optimizing_symbolic_size_warning: PyObject,
 }
 
 /// misc python utilities
@@ -80,6 +81,7 @@ pub static PY_UTILS: GILLazyPy<PyUtils> = GILLazyPy::new_py(|py| {
         tensor_from_bytes: get("tensor_from_bytes"),
         assert_tensors_close: get("assert_tensors_close"),
         builtins: builtins.into(),
+        optimizing_symbolic_size_warning: get("OptimizingSymbolicSizeWarning"),
     }
 });
 
@@ -598,3 +600,9 @@ impl<'source, T: FromPyObject<'source>> FromPyObject<'source> for MaybeNotSet<T>
         Ok(MaybeNotSet(out))
     }
 }
+
+pub fn is_python_running() -> bool {
+    // I assume this is fine?
+    // possibly not thread safe...
+    unsafe { pyo3::ffi::Py_IsInitialized() != 0 }
+}
