@@ -30,7 +30,7 @@ use crate::{
 /// OpSize is a memory optimization of Option<usize> that stores a 63-bit int and one "is this none" bit
 #[derive(Copy, Clone, PartialEq, Eq)]
 pub struct OpSize(pub u64);
-pub type OpShape = Sv<[OpSize; 6]>;
+pub type OpShape = Sv<[OpSize; 8]>;
 
 impl OpSize {
     pub const SHIFT: usize = 63;
@@ -103,7 +103,7 @@ pub fn shape_to_op_shape(shape: &Shape) -> OpShape {
 // RInnerInts distribution is heavy tailed - almost everything's 1, but there might be up to 10.
 // we use multiple of 8 bc they're u8 and we'd just pad if we used less
 pub type RInnerInts = Sv<[AxisInt; 8]>;
-pub type RInts = Sv<[RInnerInts; 8]>;
+pub type RInts = Sv<[RInnerInts; 4]>;
 
 #[pyclass]
 #[derive(Clone, PartialEq, Eq)]
@@ -929,6 +929,16 @@ impl RearrangeSpec {
         self.conform_to_input_shape(&shape)
     }
 
+    pub fn next_axis(&self) -> u8 {
+        // input_ints are included in output_ints, so we don't check them
+        self.output_ints
+            .iter()
+            .flatten()
+            .max()
+            .map(|x| x + 1)
+            .unwrap_or(0)
+    }
+
     #[pyo3(name = "expand_to")]
     fn expand_to_py(&self, shape: Shape) -> Result<RearrangeSpec> {
         self.expand_to(&shape)
