@@ -8,8 +8,6 @@ use uuid::Uuid;
 use crate::{
     opt_einsum::EinsumSpec,
     rearrange_spec::{OpSize, RearrangeSpec},
-    tensor_db::save_tensor,
-    tensor_util::{Shape, TensorAxisIndex, TensorIndex, TorchDeviceDtypeOp},
 };
 
 pub trait RustRepr {
@@ -156,20 +154,6 @@ impl RustRepr for EinsumSpec {
     }
 }
 
-impl RustRepr for TorchDeviceDtypeOp {
-    fn repr(&self) -> String {
-        if let (None, None) = (&self.device, &self.dtype) {
-            "TorchDeviceDtypeOp::default()".to_owned()
-        } else {
-            format!(
-                "TorchDeviceDtypeOp {{ device: {}, dtype: {} }}",
-                self.device.repr(),
-                self.dtype.repr(),
-            )
-        }
-    }
-}
-
 pub struct ReprWrapper(pub String);
 
 impl RustRepr for ReprWrapper {
@@ -178,66 +162,6 @@ impl RustRepr for ReprWrapper {
     }
 }
 
-impl TensorAxisIndex {
-    pub fn repr(
-        &self,
-        bound: usize,
-        device_dtype: &TorchDeviceDtypeOp,
-        tensors_as_randn: bool,
-    ) -> String {
-        match self {
-            TensorAxisIndex::Single(i) => format!("TensorAxisIndex::Single({})", i),
-            TensorAxisIndex::Tensor(t) => {
-                if tensors_as_randn {
-                    format!(
-                        "TensorAxisIndex::new_tensor_randint_seeded({}, {}, {}, {})",
-                        t.shape()[0].repr(),
-                        bound.repr(),
-                        device_dtype.repr(),
-                        t.hash_usize().unwrap(),
-                    )
-                } else {
-                    save_tensor((**t).clone(), false).unwrap();
-                    let hash_base16: &str = &t.hashed().hash_base16().unwrap();
-                    format!(
-                        "TensorAxisIndex::Tensor(tensor_from_hash({}).unwrap())",
-                        hash_base16.repr()
-                    )
-                }
-            }
-            TensorAxisIndex::Slice(s) => format!(
-                "TensorAxisIndex::Slice(Slice {{ start:{}, stop:{} }})",
-                s.start.repr(),
-                s.stop.repr()
-            ),
-        }
-    }
-}
-
-impl TensorIndex {
-    pub fn repr(
-        &self,
-        shape: Shape,
-        device_dtype: &TorchDeviceDtypeOp,
-        tensors_as_randn: bool,
-    ) -> String {
-        // shape: shape of the tensor we're indexing into
-        format!(
-            "TensorIndex ( vec![{}] )",
-            self.0
-                .iter()
-                .enumerate()
-                .map(|(i, tensor_axis_index)| tensor_axis_index.repr(
-                    shape[i],
-                    device_dtype,
-                    tensors_as_randn
-                ))
-                .collect::<Vec<String>>()
-                .join(", ")
-        )
-    }
-}
-
 #[allow(unused)]
 pub fn rustfmt_string(string: &str) -> String {
     // writing file bc dumb
