@@ -7,7 +7,7 @@ use std::{
     str::FromStr,
 };
 
-use anyhow::{bail, Context, Error, Result};
+use anyhow::{anyhow, bail, Context, Error, Result};
 use macro_rules_attribute::apply;
 use pyo3::{
     exceptions,
@@ -30,7 +30,7 @@ use crate::{
     tensor_db::{get_tensor_prefix, save_tensor},
     util::HashBytes,
 };
-pub type Shape = Sv<[usize; 6]>; // TODO: maybe this should be struct which preserves invariant that dim shapes >= 0?
+pub type Shape = Sv<[usize; 4]>; // TODO: maybe this should be struct which preserves invariant that dim shapes >= 0?
 
 #[crate::pyo3::pyfunction(shapes = "*")]
 #[pyo3(name = "broadcast_shapes")]
@@ -38,13 +38,25 @@ pub fn broadcast_shapes_py(shapes: Vec<Shape>) -> Result<PyShape> {
     broadcast_shapes(&shapes).map(PyShape)
 }
 
+/// broadcast, but don't broadcast on 1s
+pub fn right_align_shapes(shapes: &[impl Borrow<[usize]>]) -> Result<Shape> {
+    broadcast_shapes_impl(shapes, false)
+}
+
 pub fn broadcast_shapes(shapes: &[impl Borrow<[usize]>]) -> Result<Shape> {
+    broadcast_shapes_impl(shapes, true)
+}
+
+pub fn broadcast_shapes_impl(
+    shapes: &[impl Borrow<[usize]>],
+    special_case_ones: bool,
+) -> Result<Shape> {
     if shapes.is_empty() {
         return Ok(sv![]);
     }
     let ranks: Vec<usize> = shapes.iter().map(|x| x.borrow().len()).collect();
     let out_rank = *ranks.iter().max().unwrap();
-    let mut result: Shape = sv![1; out_rank];
+    let mut existing = vec![None; out_rank];
     for axis_from_end in 1..out_rank + 1 {
         for (i, shape) in shapes.iter().enumerate() {
             let axis_for_shape = ranks[i].checked_sub(axis_from_end);
@@ -52,11 +64,15 @@ pub fn broadcast_shapes(shapes: &[impl Borrow<[usize]>]) -> Result<Shape> {
                 None => {}
                 Some(j) => {
                     let axis_len = shape.borrow()[j];
-                    if axis_len == 0
-                        || axis_len != 1
-                            && result[out_rank - axis_from_end] != 1
-                            && result[out_rank - axis_from_end] != axis_len
-                    {
+                    let cur = existing[out_rank - axis_from_end];
+                    let success = cur
+                        .map(|x| {
+                            x == axis_len
+                                || (special_case_ones && x == 1 && axis_len != 0)
+                                || (special_case_ones && axis_len == 1 && x != 0)
+                        })
+                        .unwrap_or(true);
+                    if !success {
                         bail!(MiscInputError::NotBroadcastable {
                             shapes: shapes
                                 .iter()
@@ -64,13 +80,75 @@ pub fn broadcast_shapes(shapes: &[impl Borrow<[usize]>]) -> Result<Shape> {
                                 .collect(),
                         });
                     }
-                    result[out_rank - axis_from_end] =
-                        max(axis_len, result[out_rank - axis_from_end]);
+                    existing[out_rank - axis_from_end] =
+                        Some(cur.map(|x| max(x, axis_len)).unwrap_or(axis_len));
                 }
             }
         }
     }
-    Ok(result)
+    Ok(existing.into_iter().map(|x| x.unwrap()).collect())
+}
+
+#[test]
+fn test_broadcast_shapes() {
+    let run = |x: &[Shape]| broadcast_shapes(x);
+    let s = |x: Shape| x;
+    assert_eq!(run(&[]).unwrap(), Shape::new());
+    assert_eq!(run(&[sv![], sv![]]).unwrap(), s(sv![]));
+    assert_eq!(run(&[sv![], sv![]]).unwrap(), s(sv![]));
+    assert_eq!(run(&[sv![1], sv![]]).unwrap(), s(sv![1]));
+    assert_eq!(run(&[sv![], sv![1]]).unwrap(), s(sv![1]));
+    assert_eq!(run(&[sv![3], sv![]]).unwrap(), s(sv![3]));
+    assert_eq!(run(&[sv![], sv![3]]).unwrap(), s(sv![3]));
+    assert_eq!(run(&[sv![1], sv![1]]).unwrap(), s(sv![1]));
+    assert_eq!(run(&[sv![3], sv![3]]).unwrap(), s(sv![3]));
+    assert_eq!(run(&[sv![1], sv![4, 3]]).unwrap(), s(sv![4, 3]));
+    assert_eq!(run(&[sv![4, 3], sv![4, 1]]).unwrap(), s(sv![4, 3]));
+    assert_eq!(run(&[sv![4, 3], sv![1, 3]]).unwrap(), s(sv![4, 3]));
+    assert_eq!(run(&[sv![4, 1], sv![4, 3]]).unwrap(), s(sv![4, 3]));
+    assert_eq!(run(&[sv![1, 3], sv![4, 3]]).unwrap(), s(sv![4, 3]));
+    assert_eq!(
+        run(&[sv![1, 3], sv![4, 3], sv![5, 4, 1]]).unwrap(),
+        s(sv![5, 4, 3])
+    );
+    assert_eq!(
+        run(&[sv![3, 4, 5], sv![3, 4, 5], sv![3, 4, 5]]).unwrap(),
+        s(sv![3, 4, 5])
+    );
+
+    assert!(run(&[sv![5], sv![6]]).is_err());
+    assert!(run(&[sv![3, 5], sv![5, 1]]).is_err());
+    assert!(run(&[sv![5, 0], sv![5, 1]]).is_err());
+    assert!(run(&[sv![5, 2], sv![5, 1]]).is_ok());
+    assert!(run(&[sv![1, 2, 2], sv![1, 2], sv![3, 2, 2, 2, 172112]]).is_err());
+
+    let run_align = |x: &[Shape]| right_align_shapes(x);
+
+    assert_eq!(run_align(&[]).unwrap(), Shape::new());
+    assert_eq!(run_align(&[sv![], sv![]]).unwrap(), s(sv![]));
+    assert_eq!(run_align(&[sv![], sv![]]).unwrap(), s(sv![]));
+    assert_eq!(run_align(&[sv![1], sv![]]).unwrap(), s(sv![1]));
+    assert_eq!(run_align(&[sv![], sv![1]]).unwrap(), s(sv![1]));
+    assert_eq!(run_align(&[sv![3], sv![]]).unwrap(), s(sv![3]));
+    assert_eq!(run_align(&[sv![], sv![3]]).unwrap(), s(sv![3]));
+    assert_eq!(run_align(&[sv![1], sv![1]]).unwrap(), s(sv![1]));
+    assert_eq!(run_align(&[sv![3], sv![3]]).unwrap(), s(sv![3]));
+    assert!(run_align(&[sv![1], sv![4, 3]]).is_err());
+    assert!(run_align(&[sv![4, 3], sv![4, 1]]).is_err());
+    assert!(run_align(&[sv![4, 3], sv![1, 3]]).is_err());
+    assert!(run_align(&[sv![4, 1], sv![4, 3]]).is_err());
+    assert!(run_align(&[sv![1, 3], sv![4, 3]]).is_err());
+    assert!(run_align(&[sv![1, 3], sv![4, 3], sv![5, 4, 1]]).is_err());
+    assert_eq!(
+        run_align(&[sv![3, 4, 5], sv![3, 4, 5], sv![3, 4, 5]]).unwrap(),
+        s(sv![3, 4, 5])
+    );
+
+    assert!(run_align(&[sv![5], sv![6]]).is_err());
+    assert!(run_align(&[sv![3, 5], sv![5, 1]]).is_err());
+    assert!(run_align(&[sv![5, 0], sv![5, 1]]).is_err());
+    assert!(run_align(&[sv![5, 2], sv![5, 1]]).is_err());
+    assert!(run_align(&[sv![1, 2, 2], sv![1, 2], sv![3, 2, 2, 2, 172112]]).is_err());
 }
 
 #[apply(python_error_exception)]
@@ -86,14 +164,14 @@ pub enum MiscInputError {
 
     #[error("Children multiple dtypes {a:?} {b:?} ({e_name})")]
     ChildrenMultipleDtypes {
-        a: Option<String>,
-        b: Option<String>,
+        a: Option<TorchDtype>,
+        b: Option<TorchDtype>,
     },
 
     #[error("Children multiple dtypes {a:?} {b:?} ({e_name})")]
     ChildrenMultipleDevices {
-        a: Option<String>,
-        b: Option<String>,
+        a: Option<TorchDevice>,
+        b: Option<TorchDevice>,
     },
 }
 
@@ -435,7 +513,7 @@ impl TensorAxisIndex {
                 .unwrap();
             let mut kwargs = HashMap::default();
             if let Some(dtype) = device_dtype.dtype {
-                let dtype: &str = &dtype;
+                let dtype: &str = &String::from(&dtype);
                 kwargs.insert("dtype", PY_UTILS.torch.getattr(py, dtype).unwrap());
             }
             if let Some(device) = device_dtype.device {
@@ -854,9 +932,9 @@ pub fn string_escape(s: &String) -> String {
 #[derive(Debug, Clone, PartialEq, Eq, Default)]
 pub struct TorchDeviceDtypeOp {
     #[pyo3(get)]
-    pub device: Option<String>,
+    pub device: Option<TorchDevice>,
     #[pyo3(get)]
-    pub dtype: Option<String>,
+    pub dtype: Option<TorchDtype>,
 }
 
 #[pymethods]
@@ -869,7 +947,7 @@ impl TorchDeviceDtypeOp {
 
     #[new]
     #[args(device = "None", dtype = "None")]
-    fn new(device: Option<String>, dtype: Option<String>) -> Self {
+    fn new(device: Option<TorchDevice>, dtype: Option<TorchDtype>) -> Self {
         Self { device, dtype }
     }
 
@@ -878,25 +956,120 @@ impl TorchDeviceDtypeOp {
             "TorchDeviceDtypeOp(device={}, dtype={})",
             self.device
                 .as_ref()
-                .map_or("None".to_owned(), string_escape),
-            self.dtype.as_ref().map_or("None".to_owned(), string_escape)
+                .map(|x| format!("\"{}\"", String::from(x)))
+                .unwrap_or("None".to_owned()),
+            self.dtype
+                .as_ref()
+                .map(|x| format!("\"{}\"", String::from(x)))
+                .unwrap_or("None".to_owned())
         )
     }
 }
+macro_rules! lit_or {
+    ($a:ident, $b:literal) => {
+        $b
+    };
+    ($a:ident) => {
+        stringify!($a)
+    };
+}
+macro_rules! string_enum{
+    {$name:ident, ($($var:ident $(+$lit:literal)?),*)}=>{
+        #[derive(Copy, Clone, Debug, Hash, PartialEq, Eq, PartialOrd, Ord)]
+        #[allow(non_camel_case_types)]
+        pub enum $name{
+            $($var,)*
+        }
+        impl From<&$name> for String{
+            #[allow(unreachable_code)]
+            fn from(value:&$name)->Self{
+                match value{
+                    $($name::$var=>{
+                        $(return $lit.to_owned();)*
+                        return stringify!($var).to_owned();
+                    }),*
+                }
+            }
+        }
+        impl TryFrom<String> for $name{
+            type Error = Error;
+            fn try_from(value:String)->Result<$name>{
+                let inp_str_str:&str = &value;
+                match inp_str_str{
+                    $(
+                        lit_or!($var $(,$lit)?)=>Ok($name::$var),
+                    )*
+                    _=>Err(anyhow!("invalid value for enum {}, {}", stringify!($name), value).into())
+                }
+            }
+        }
+        impl IntoPy<PyObject> for $name {
+            fn into_py(self, py: Python<'_>) -> PyObject {
+                String::from(&self).into_py(py)
+            }
+        }
+        impl<'source> FromPyObject<'source> for $name {
+            fn extract(circuit_obj: &'source PyAny) -> PyResult<Self> {
+                let inp_str:String = circuit_obj.extract()?;
+                let resulty:Result<$name> = inp_str.try_into();
+                resulty.map_err(|e|e.into())
+            }
+        }
+    }
+}
+
+string_enum! {
+    TorchDtype,
+    (
+        float16,
+        bfloat16,
+        float32,
+        float64,
+        int8,
+        int16,
+        int32,
+        int64,
+        uint8,
+        bool
+    )
+}
+
+string_enum! {
+    TorchDevice,
+    (
+        Cpu+"cpu",
+        Cuda0+"cuda:0",
+        Cuda1+"cuda:1",
+        Cuda2+"cuda:2",
+        Cuda3+"cuda:3",
+        Cuda4+"cuda:4",
+        Cuda5+"cuda:5",
+        Cuda6+"cuda:6",
+        Cuda7+"cuda:7",
+        Cuda8+"cuda:8",
+        Cuda9+"cuda:9",
+        Cuda10+"cuda:10",
+        Cuda11+"cuda:11",
+        Cuda12+"cuda:12",
+        Cuda13+"cuda:13",
+        Cuda14+"cuda:14",
+        Cuda15+"cuda:15"
+    )
+}
 
 #[pyclass]
 #[derive(Debug, Clone, PartialEq, Eq)]
 pub struct TorchDeviceDtype {
     #[pyo3(get)]
-    pub device: String,
+    pub device: TorchDevice,
     #[pyo3(get)]
-    pub dtype: String,
+    pub dtype: TorchDtype,
 }
 
 #[pymethods]
 impl TorchDeviceDtype {
     #[new]
-    fn new(device: String, dtype: String) -> Self {
+    fn new(device: TorchDevice, dtype: TorchDtype) -> Self {
         Self { device, dtype }
     }
     #[staticmethod]
@@ -922,9 +1095,9 @@ impl TorchDeviceDtype {
 
     pub fn __repr__(&self) -> String {
         format!(
-            "TorchDeviceDtype(device={}, dtype={})",
-            string_escape(&self.device),
-            string_escape(&self.dtype)
+            "TorchDeviceDtype(device=\"{}\", dtype=\"{}\")",
+            String::from(&self.device),
+            String::from(&self.dtype),
         )
     }
 }
@@ -952,18 +1125,20 @@ impl TorchDeviceDtype {
                 .unwrap()
                 .extract::<String>(py)
                 .unwrap()[6..]
-                .to_owned(),
+                .to_owned()
+                .try_into()
+                .unwrap(),
         })
     }
     pub fn size(&self) -> usize {
         let map = HashMap::from_iter([
-            ("float32".to_owned(), 4),
-            ("float64".to_owned(), 8),
-            ("float16".to_owned(), 2),
-            ("int64".to_owned(), 8),
-            ("int32".to_owned(), 4),
-            ("int16".to_owned(), 2),
-            ("int8".to_owned(), 1),
+            (TorchDtype::float32, 4),
+            (TorchDtype::float64, 8),
+            (TorchDtype::float16, 2),
+            (TorchDtype::int64, 8),
+            (TorchDtype::int32, 4),
+            (TorchDtype::int16, 2),
+            (TorchDtype::int8, 1),
         ]);
         map[&self.dtype]
     }
@@ -1010,8 +1185,8 @@ impl TorchDeviceDtypeOp {
 
     pub fn unwrap_or_defaults(self) -> TorchDeviceDtype {
         TorchDeviceDtype {
-            dtype: self.dtype.unwrap_or_else(|| "float32".to_owned()),
-            device: self.device.unwrap_or_else(|| "cpu".to_owned()),
+            dtype: self.dtype.unwrap_or_else(|| TorchDtype::float32),
+            device: self.device.unwrap_or_else(|| TorchDevice::Cpu),
         }
     }
 
@@ -1040,7 +1215,6 @@ impl Default for TorchDeviceDtype {
     }
 }
 
-// TODO: break up and move part to parsing.rs
 #[apply(python_error_exception)]
 #[base_error_name(Parse)]
 #[base_exception(PyValueError)]
@@ -1049,9 +1223,6 @@ pub enum ParseError {
     #[error("Parsing number failed string='{string}', err={err_string} ({e_name})")]
     InvalidNumber { string: String, err_string: String },
 
-    #[error("extra='{extra}' ({e_name})")]
-    ExpectedNoExtraInfo { extra: String },
-
     #[error("string='{string}'\nindex should look like '[1,3:5,f6d587856c8444a9]'\n({e_name})")]
     InvalidIndex { string: String },
 
