@@ -654,7 +654,7 @@ macro_rules! fn_struct{
                     }
                 }
             }
-            impl Debug for [< $name DynStruct >]{
+            impl std::fmt::Debug for [< $name DynStruct >]{
                 fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
                     write!(f,"dyn_fn_struct")
                 }
@@ -685,3 +685,26 @@ pub fn transpose<T>(v: Vec<Vec<T>>, inner_len: usize) -> Vec<Vec<T>> {
         })
         .collect()
 }
+
+pub struct DimNumMaker {
+    pub running: usize,
+}
+
+impl Default for DimNumMaker {
+    fn default() -> Self {
+        Self { running: 0 }
+    }
+}
+
+impl DimNumMaker {
+    pub fn next(&mut self) -> usize {
+        let out = self.running;
+        self.running += 1;
+        out
+    }
+    pub fn next_range(&mut self, len: usize) -> std::ops::Range<usize> {
+        let out = self.running..self.running + len;
+        self.running += len;
+        out
+    }
+}
