@@ -139,6 +139,7 @@ fn rust_circuit(py: Python<'_>, m: &PyModule) -> PyResult<()> {
     m.add_class::<get_update_node::Expander>()?;
     m.add_class::<circuit_base::print::PrintOptions>()?;
     m.add_class::<circuit_base::print_html::PrintHtmlOptions>()?;
+    m.add_class::<circuit_base::set_of_circuits::SetOfCircuitIdentities>()?;
     m.add_function(wrap_pyfunction!(
         circuit_base::print::set_debug_print_options,
         m
@@ -369,6 +370,20 @@ fn rust_circuit(py: Python<'_>, m: &PyModule) -> PyResult<()> {
         circuit_base::module::conform_all_modules,
         m
     )?)?;
+    m.add_function(wrap_pyfunction!(
+        circuit_base::module::get_children_with_symbolic_sizes,
+        m
+    )?)?;
+    m.add_function(wrap_pyfunction!(
+        circuit_base::module::any_children_with_symbolic_sizes,
+        m
+    )?)?;
+    m.add(
+        "OptimizingSymbolicSizeWarning",
+        rr_util::py_types::PY_UTILS
+            .optimizing_symbolic_size_warning
+            .clone(),
+    )?;
     m.add_function(wrap_pyfunction!(
         circuit_base::expand_node::replace_expand_bottom_up_dict_py,
         m
@@ -422,6 +437,22 @@ fn rust_circuit(py: Python<'_>, m: &PyModule) -> PyResult<()> {
     m.add_class::<circuit_base::parsing::Parser>()?;
 
     m.add_function(wrap_pyfunction!(circuit_base::print_circuit_type_check, m)?)?;
+    m.add_function(wrap_pyfunction!(
+        circuit_rewrites::module_rewrite::elim_empty_module,
+        m
+    )?)?;
+    m.add_function(wrap_pyfunction!(
+        circuit_rewrites::module_rewrite::elim_no_input_module,
+        m
+    )?)?;
+    m.add_function(wrap_pyfunction!(
+        circuit_rewrites::module_rewrite::module_remove_unused_inputs,
+        m
+    )?)?;
+    m.add_function(wrap_pyfunction!(
+        circuit_rewrites::module_rewrite::py_deep_module_remove_unused_inputs,
+        m
+    )?)?;
     m.add_function(wrap_pyfunction!(
         circuit_rewrites::module_rewrite::extract_rewrite_raw,
         m
@@ -465,6 +496,19 @@ fn rust_circuit(py: Python<'_>, m: &PyModule) -> PyResult<()> {
         m
     )?)?;
     m.add_class::<nb_operations::modules::BindItem>()?;
+    m.add_class::<nb_operations::modules::ModulePusher>()?;
+    m.add_function(wrap_pyfunction!(
+        nb_operations::modules::default_nested_module_namer,
+        m
+    )?)?;
+    m.add_function(wrap_pyfunction!(
+        nb_operations::modules::extract_symbols,
+        m
+    )?)?;
+    m.add_function(wrap_pyfunction!(
+        nb_operations::modules::extract_symbols_get,
+        m
+    )?)?;
 
     // create dummy object for all the type aliases we use in the stub file
     // TODO: maybe these should somehow be the actual types we use in the stub file
@@ -492,10 +536,24 @@ fn rust_circuit(py: Python<'_>, m: &PyModule) -> PyResult<()> {
         "Binder",
         "UpdatedIterativeMatcher",
         "UpdatedIterativeMatcherIn",
+        "PrintOptionsBase",
+        "CliColor",
+        "CircuitColorer",
+        "CircuitHtmlColorer",
+        "NestedModuleNamer",
     ] {
         m.add(dummy_name, py.None())?;
     }
 
+    m.add_function(wrap_pyfunction!(
+        nb_operations::cumulant_rewrites::rewrite_cum_to_circuit_of_cum,
+        m
+    )?)?;
+    m.add_function(wrap_pyfunction!(
+        nb_operations::cumulant_rewrites::kappa_term_py,
+        m
+    )?)?;
+
     error::register_exceptions(py, m)?;
 
     Ok(())
